/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include "codec_vfw.h"
#include <stdio.h>
#include <memory.h>

#include "inttypes.h"
#include "config.h"
#include "encode.h"
#include "decode.h"

#include "resource.h"


#define ABS(x) ( (x)<0 ? (-(x)) : (x) )


typedef struct {
	void *hndEnc;
	int EncConfig_id;
	DWORD biSizeImage;
	CRITICAL_SECTION csLockEnc;

	void *hndDec;
	int DecConfig_id;
	CRITICAL_SECTION csLockDec;
} driver_context_t;



HMODULE hmodule_cedocida = NULL;
configvars_t CurrentConfiguration;
volatile int LockConfig = 0;	//we use this for atomic access to the global configuration



// helper class to lock a critical section, and unlock it automatically
class CAutoLockCS 
{
protected:
    CRITICAL_SECTION *m_pLock;

public:
	CAutoLockCS(CRITICAL_SECTION *plock)
	{
		m_pLock = plock;
		EnterCriticalSection(m_pLock);
	};

	~CAutoLockCS() 
	{ 
		LeaveCriticalSection(m_pLock);	
	};
};




__declspec(dllexport) LRESULT WINAPI DriverProc(DWORD dwDriverID, HDRVR hDriver, UINT uiMessage, LPARAM lParam1, LPARAM lParam2) 
{
	driver_context_t * hndDriver = (driver_context_t*)dwDriverID; 

	switch (uiMessage) {

		/****************************************
			standard driver messages
		****************************************/
		case DRV_INSTALL: return (LRESULT)DRV_OK;
		case DRV_REMOVE: return (LRESULT)DRV_OK;
		case DRV_ENABLE: return (LRESULT)1L;
		case DRV_DISABLE: return 0;
		case DRV_LOAD: return my_drv_load();
		case DRV_OPEN: return my_drv_open(lParam2);
		case DRV_CLOSE: return my_drv_close( hndDriver );
		case DRV_FREE: return (LRESULT)1L;
		case DRV_QUERYCONFIGURE: return (LRESULT)1L;
		case DRV_CONFIGURE: return my_drv_configure(lParam1, lParam2);

		/****************************************
			state messages
		****************************************/
		case ICM_ABOUT: return my_icm_about(lParam1,lParam2);
		case ICM_CONFIGURE: return my_icm_configure(lParam1,lParam2);
		case ICM_GETSTATE: return my_icm_getstate(lParam1, lParam2);
		case ICM_SETSTATE: return my_icm_setstate(lParam1, lParam2);
		case ICM_GET: return 0;
		case ICM_GETQUALITY:
		case ICM_SETQUALITY:
		case ICM_GETDEFAULTQUALITY:
		case ICM_GETDEFAULTKEYFRAMERATE: return ICERR_UNSUPPORTED;
		case ICM_GETINFO: return my_icm_getinfo(lParam1, lParam2);

		/****************************************
			compress messages
		****************************************/
		case ICM_COMPRESS_QUERY: return my_icm_compress_query(lParam1, lParam2);
		case ICM_COMPRESS_GET_FORMAT: return my_icm_compress_get_format(lParam1, lParam2);
		case ICM_COMPRESS_GET_SIZE: return my_icm_get_size(lParam1);
		case ICM_COMPRESS_BEGIN: return my_icm_compress_begin(hndDriver, lParam1, lParam2);
		case ICM_COMPRESS: return my_icm_compress(hndDriver, lParam1, lParam2);
		case ICM_COMPRESS_END: return my_icm_compress_end(hndDriver, lParam1, lParam2);

		/****************************************
			decompress messages
		****************************************/
		case ICM_DECOMPRESS_GET_PALETTE: return ICERR_BADFORMAT;
		case ICM_DECOMPRESS_SET_PALETTE: return ICERR_UNSUPPORTED;
		case ICM_DECOMPRESS_QUERY: return my_icm_decompress_query(lParam1, lParam2);
		case ICM_DECOMPRESS_GET_FORMAT: return my_icm_decompress_get_format(lParam1, lParam2);
		case ICM_DECOMPRESS_BEGIN: return my_icm_decompress_begin(hndDriver, lParam1, lParam2);
		case ICM_DECOMPRESS: return my_icm_decompress(hndDriver, lParam1, lParam2);
		case ICM_DECOMPRESS_END: return my_icm_decompress_end(hndDriver, lParam1, lParam2);
	}

	if (uiMessage < DRV_USER)
		return DefDriverProc(dwDriverID, hDriver, uiMessage, lParam1, lParam2);
	else
		return ICERR_UNSUPPORTED;
}


BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID) 
{
	if (fdwReason==DLL_PROCESS_ATTACH)
	{
		hmodule_cedocida = (HMODULE)hinstDLL;
	}
	
	return TRUE;
}



/****************************************
	helper functions
****************************************/

void get_configuration( configvars_t *lpmyconfig )
{
	int value;
	int nTry=100;

	for (;;)
	{
		value = InterlockedCompareExchange( (LONG*)(&LockConfig), 1, 0 );
		if (value==0)
		{
			memcpy( lpmyconfig, &CurrentConfiguration, sizeof(configvars_t) );
			InterlockedExchange( (LONG*)(&LockConfig), 0 );

			return;
		}

		nTry--;
		if (nTry<=0)
			break;

		Sleep(1);
	}
}


void set_configuration( configvars_t *lpmyconfig )
{
	int value;
	int nTry=100;

	for (;;)
	{
		value = InterlockedCompareExchange( (LONG*)(&LockConfig), 1, 0 );
		if (value==0)
		{
			int id = CurrentConfiguration.id + 1;		
			memcpy( &CurrentConfiguration, lpmyconfig, sizeof(configvars_t) );
			CurrentConfiguration.id = id;
			InterlockedExchange( (LONG*)(&LockConfig), 0 );

			return;
		}

		nTry--;
		if (nTry<=0)
			break;

		Sleep(1);
	}
}


static bool is_DVSD(LPBITMAPINFOHEADER lpbiIn)
{
	if (!lpbiIn) 
		return 0;

	if (  (lpbiIn->biCompression == FOURCC_DVSD) ||
		  (lpbiIn->biCompression == FOURCC_dvsd)  )
		  return TRUE;

	return FALSE;
}


int get_colorformat(LPBITMAPINFOHEADER lpbiIn)
{
	int fourcc;
	int bitcount;

	if (!lpbiIn) 
		return 0;

	fourcc = lpbiIn->biCompression;
	bitcount = lpbiIn->biBitCount;

	if (  ( fourcc == 0 || fourcc == ' BID' ) && ( bitcount == 32)  ) 
		return COLORFORMAT_RGB32;
	else if (  ( fourcc == 0 || fourcc == ' BID' ) && ( bitcount == 24)  ) 
		return COLORFORMAT_RGB24;
	else if ( fourcc == FOURCC_YUY2 ) 
		return COLORFORMAT_YUY2;
	else if ( fourcc == FOURCC_YV12 ) 
		return COLORFORMAT_YV12;
	else
		return COLORFORMAT_UNKNOWN;
}


int get_input_colorformat(LPBITMAPINFOHEADER lpbiIn)
{
	configvars_t lpmyconfig[1];
	get_configuration( lpmyconfig );

	int color_format = get_colorformat(lpbiIn);

	if (color_format<0)
		return color_format;

	color_format &= lpmyconfig->encoder_input_format;

	if (color_format==0)
		return COLORFORMAT_NA;

	return color_format;
}


int get_output_colorformat(LPBITMAPINFOHEADER lpbiIn)
{
	configvars_t lpmyconfig[1];
	get_configuration( lpmyconfig );

	int color_format = get_colorformat(lpbiIn);

	if (color_format<0)
		return color_format;

	color_format &= lpmyconfig->decoder_output_format;

	if (color_format==0)
		return COLORFORMAT_NA;

	return color_format;
}


/****************************************
	standard driver messages
****************************************/

INT_PTR CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

LONG my_drv_configure(LPARAM lParam1, LPARAM lParam2)
{
	HWND hwnd = (HWND)lParam1;

	configvars_t lpmyconfig[1];
	get_configuration( lpmyconfig );

	DialogBoxParam(
		hmodule_cedocida, 
		MAKEINTRESOURCE(IDD_DIALOG_CONFIGURE), 
		hwnd, 
		ConfigureDialogProc, 
		(LPARAM)(&(lpmyconfig[0])) );

	set_configuration( lpmyconfig );

	return DRV_OK;
}


LONG my_drv_load() 
{
	configvars_t lpmyconfig[1];
	load_configuration( lpmyconfig );
	set_configuration( lpmyconfig );

	return (LRESULT)1L;
}


LONG my_drv_open(LPARAM lParam2) 
{
	ICOPEN *icinfo = (ICOPEN*)lParam2;

	driver_context_t * hndDriver = (driver_context_t*)malloc( sizeof(driver_context_t) );
	if (hndDriver==NULL)
		return NULL;

	if (icinfo)
	{
		if ( icinfo->fccType!=ICTYPE_VIDEO ) 
		{
			icinfo->dwError = ICERR_UNSUPPORTED;

			free( hndDriver );
			return NULL;
		}

		icinfo->dwError = ICERR_OK;
	}

	memset( hndDriver, 0x00, sizeof(driver_context_t) );

	hndDriver->EncConfig_id = CurrentConfiguration.id - 1;
	hndDriver->DecConfig_id = CurrentConfiguration.id - 1;

	InitializeCriticalSection( &hndDriver->csLockEnc );
	InitializeCriticalSection( &hndDriver->csLockDec );

	return (DWORD)hndDriver;
}


LONG my_drv_close( void *hnd ) 
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;

	if (hndDriver)
	{
		if ( hndDriver->hndEnc )
			encoder_close( hndDriver->hndEnc );

		if ( hndDriver->hndDec )
			decoder_close( hndDriver->hndDec );

		DeleteCriticalSection( &hndDriver->csLockEnc );
		DeleteCriticalSection( &hndDriver->csLockDec );

		free(hndDriver);
	}

	return (LRESULT)1L;
}


/****************************************
	state messages
****************************************/

LONG my_icm_about(LPARAM lParam1, LPARAM lParam2)
{
	return ICERR_UNSUPPORTED;
}


LONG my_icm_configure(LPARAM lParam1, LPARAM lParam2)
{
	if (lParam1 == -1)
		return ICERR_OK;
	else 
		my_drv_configure(lParam1, lParam2);

	return ICERR_OK;
}


LONG my_icm_getstate(LPARAM lParam1, LPARAM lParam2)
{
	configvars_t * pconfig = (configvars_t *)(void*)lParam1;
	int nbytes = (int)lParam2;

	if (pconfig == NULL)
		return sizeof(configvars_t);

	if (nbytes != sizeof(configvars_t))
		return ICERR_BADPARAM;

	get_configuration(pconfig);
	
	return ICERR_OK;
}


LONG my_icm_setstate(LPARAM lParam1, LPARAM lParam2)
{
	configvars_t * pconfig = (configvars_t *)(void*)lParam1;
	int nbytes = (int)lParam2;

	if (pconfig == NULL)
	{
		configvars_t config;
		load_configuration(&config);
		set_configuration(&config);

		return ICERR_OK;
	}

	if (nbytes != sizeof(configvars_t))
		return ICERR_BADPARAM;

	set_configuration(pconfig);

	return ICERR_OK;
}


LONG my_icm_getinfo(LPARAM lParam1, LPARAM lParam2)
{
	ICINFO* icinfo = (ICINFO *)lParam1;
	DWORD dwSize = (DWORD)lParam2;

	if (icinfo == NULL) 
		return sizeof(ICINFO);

	if (dwSize < sizeof(ICINFO)) 
		return 0;

	icinfo->dwSize            = sizeof(ICINFO);
	icinfo->fccType           = ICTYPE_VIDEO;
	icinfo->fccHandler        = FOURCC_dvsd;
	icinfo->dwFlags           = 0;
	icinfo->dwVersion         = 0;
	icinfo->dwVersionICM      = ICVERSION;
	wcscpy_s(icinfo->szName, 16, CODEC_NAME); 
	wcscpy_s(icinfo->szDescription, 128, CODEC_DESCRIPTION);

	return (DWORD)lParam2;
}


/****************************************
	compress messages
****************************************/

LONG my_icm_compress_query(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;
	
	if (!lpbiIn)
		return ICERR_BADFORMAT;

	if ( get_input_colorformat(lpbiIn) < 0 )
		return ICERR_BADFORMAT;

	if (!lpbiOut) 
		return ICERR_OK;

	if ( !is_DVSD(lpbiOut) )
		return ICERR_BADFORMAT;
		
	return ICERR_OK;
}


LONG my_icm_compress_get_format(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if (!lpbiIn)
		return ICERR_BADFORMAT;

	if (!lpbiOut) 
		return sizeof(BITMAPINFOHEADER);

	if (ABS(lpbiIn->biHeight)<576)
	{
		lpbiOut->biHeight = 480;
		lpbiOut->biSizeImage = 120000;
	}
	else
	{
		lpbiOut->biHeight = 576;
		lpbiOut->biSizeImage = 144000;
	}

	lpbiOut->biSize = sizeof(BITMAPINFOHEADER);
	lpbiOut->biWidth = 720;
	lpbiOut->biPlanes = 1;
	lpbiOut->biBitCount = 24;
	lpbiOut->biCompression = FOURCC_dvsd;
    lpbiOut->biXPelsPerMeter = 0; //117*100; 
    lpbiOut->biYPelsPerMeter = 0; //128*100; 
    lpbiOut->biClrUsed = 0; 
    lpbiOut->biClrImportant = 0; 

	return ICERR_OK;
}


LONG my_icm_get_size(LPARAM lParam1)
{
	LPBITMAPINFOHEADER const lpbiIn = (LPBITMAPINFOHEADER)lParam1;

	// 1 DV-Frame = 12 (or 10 for NTSC) DIF-Sequence  150 DIF-Blocks  80 Bytes = 144000 Bytes (or 12000 for NTSC)

	DWORD dwSize = 150*80*12;

	if (lpbiIn)
		dwSize = 150*80*( ABS(lpbiIn->biHeight)>=576 ? 12 : 10 );

	return dwSize;
}


LONG my_icm_compress_begin(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if ( (hndDriver==NULL) || (lpbiIn==NULL) || (lpbiOut==NULL) )
		return ICERR_BADFORMAT;

	if (  !( (lpbiIn->biWidth==720) && 
		     (is_DVSD(lpbiOut)) )  )
		return ICERR_BADFORMAT;

	CAutoLockCS LockObj(&hndDriver->csLockEnc);

	int is_PAL;

	if (ABS(lpbiIn->biHeight)==576) 
	{
		is_PAL = 1;
		hndDriver->biSizeImage = 144000;
	}
	else if (ABS(lpbiIn->biHeight)==480) 
	{
		is_PAL = 0;
		hndDriver->biSizeImage = 120000;
	}
	else
		return ICERR_BADFORMAT;

	int color_format = get_input_colorformat(lpbiIn);
	if (color_format<=0)
		return ICERR_BADFORMAT;

	int flip_RGB = ( lpbiIn->biHeight < 0 );	// flip_RGB==1 means first line in memory is top line of RGB-image

	if (hndDriver->hndEnc==NULL)
		hndDriver->hndEnc = encoder_open();

	if (hndDriver->hndEnc)
	{
		configvars_t lpmyconfig[1];
		get_configuration( lpmyconfig );

		encoder_configure( hndDriver->hndEnc, color_format, flip_RGB, is_PAL, lpmyconfig );
		hndDriver->EncConfig_id = lpmyconfig->id;

		return ICERR_OK;
	}

	return ICERR_BADFORMAT;
}


LONG my_icm_compress(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;
	ICCOMPRESS *icinfo = (ICCOMPRESS *)lParam1;
	DWORD dwSize = (DWORD)lParam2;
	
	if ( (hndDriver==NULL) || (icinfo==NULL) )
		return ICERR_BADPARAM;

	CAutoLockCS LockObj(&hndDriver->csLockEnc);

	if (hndDriver->hndEnc==NULL)
		return ICERR_BADPARAM;

	if (dwSize<sizeof(ICCOMPRESS))
		return ICERR_BADPARAM;

	if (icinfo->lpckid) 
		*icinfo->lpckid = FOURCC_dvsd;
	
	if (icinfo->lpdwFlags)
		*icinfo->lpdwFlags = AVIIF_KEYFRAME;

	unsigned int *in = (unsigned int *)icinfo->lpInput;
	unsigned int *out = (unsigned int *)icinfo->lpOutput;

	if ( (in==NULL) || (out==NULL) )
		return ICERR_BADFORMAT;

	if ( hndDriver->EncConfig_id != CurrentConfiguration.id )
	{
		if ( my_icm_compress_begin( hndDriver, (LPARAM)(icinfo->lpbiInput), (LPARAM)(icinfo->lpbiOutput) ) != ICERR_OK )
			return ICERR_BADFORMAT;
	}

	encode_DVSD( hndDriver->hndEnc, in, out );
	icinfo->lpbiOutput->biSizeImage = hndDriver->biSizeImage;

	return ICERR_OK;
}


LONG my_icm_compress_end(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;

	if (hndDriver)
	{
		CAutoLockCS LockObj(&hndDriver->csLockEnc);

		if (hndDriver->hndEnc)
		{
			encoder_close(hndDriver->hndEnc);
			hndDriver->hndEnc = NULL;
		}
	}

	return ICERR_OK;
}


/****************************************
	decompress messages
****************************************/

LONG my_icm_decompress_query(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;
	
	if (!lpbiIn) 
		return ICERR_BADFORMAT;

	if ( !is_DVSD(lpbiIn) )
		return ICERR_BADFORMAT;

	if (!lpbiOut) 
		return ICERR_OK;
	
	if ( get_output_colorformat(lpbiOut) < 0 ) 
		return ICERR_BADFORMAT;
		
	return ICERR_OK;
}


LONG my_icm_decompress_get_format(LPARAM lParam1, LPARAM lParam2)
{
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if (lpbiIn==NULL)
		return ICERR_BADFORMAT;

	if (lpbiOut == NULL) 
		return sizeof(BITMAPINFOHEADER);

	configvars_t lpmyconfig[1];
	get_configuration( lpmyconfig );

	if (!is_DVSD(lpbiIn)) 
		return ICERR_BADFORMAT;

	memset( lpbiOut, 0x00, sizeof(BITMAPINFOHEADER) );

	lpbiOut->biSize = sizeof(BITMAPINFOHEADER);
	lpbiOut->biWidth = 720;
	lpbiOut->biPlanes = 1;

	if ( ABS(lpbiIn->biHeight) < 576 )
		lpbiOut->biHeight = 480;
	else
		lpbiOut->biHeight = 576;

	if ( lpmyconfig->decoder_output_format & COLORFORMAT_RGB32 )
	{   // RGB32
		lpbiOut->biBitCount = 32;    
		lpbiOut->biCompression = BI_RGB;
		lpbiOut->biSizeImage = 720 * lpbiOut->biHeight * 4;

		return ICERR_OK;
	}

	if ( lpmyconfig->decoder_output_format & COLORFORMAT_RGB24 )
	{   // RGB24
		lpbiOut->biBitCount = 24;    
		lpbiOut->biCompression = BI_RGB;
		lpbiOut->biSizeImage = 720 * lpbiOut->biHeight * 3;

		return ICERR_OK;
	}

	if ( lpmyconfig->decoder_output_format & COLORFORMAT_YV12 )
	{	// YV12
		lpbiOut->biBitCount = 12;       
		lpbiOut->biCompression = FOURCC_YV12;
		lpbiOut->biSizeImage = (720 * lpbiOut->biHeight * 3)/2;

		return ICERR_OK;
	}

	if ( lpmyconfig->decoder_output_format & COLORFORMAT_YUY2 )
	{	// YUY2
		lpbiOut->biBitCount = 16;       // YUY2
		lpbiOut->biCompression = FOURCC_YUY2;
		lpbiOut->biSizeImage = 720 * lpbiOut->biHeight * 2;

		return ICERR_OK;
	}

	// RGB24
	lpbiOut->biBitCount = 24;    
	lpbiOut->biCompression = BI_RGB;
	lpbiOut->biSizeImage = 720 * lpbiOut->biHeight * 3;

	return ICERR_OK;
}


LONG my_icm_decompress_begin(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;
	LPBITMAPINFOHEADER lpbiIn = (LPBITMAPINFOHEADER)lParam1;
	LPBITMAPINFOHEADER lpbiOut = (LPBITMAPINFOHEADER)lParam2;

	if ( (hndDriver==NULL) || (lpbiIn==NULL) || (lpbiOut==NULL) )
		return ICERR_BADFORMAT;

	if ( (!is_DVSD(lpbiIn)) ||
		  (lpbiOut->biWidth!=720) )
		return ICERR_BADFORMAT;

	CAutoLockCS LockObj(&hndDriver->csLockDec);

	int is_PAL;

	if (ABS(lpbiOut->biHeight)==576)
	{
		if (lpbiIn->biSizeImage<144000) 
			return ICERR_BADFORMAT;

		is_PAL = 1;
	}
	else if (ABS(lpbiOut->biHeight)==480)
	{
		if (lpbiIn->biSizeImage<120000)
			return ICERR_BADFORMAT;

		is_PAL = 0;
	}
	else
		return ICERR_BADFORMAT;

	int color_format = get_output_colorformat(lpbiOut);
	if (color_format<=0)
		return ICERR_BADFORMAT;

	int flip_RGB = ( lpbiOut->biHeight < 0 );	// flip_RGB==1 means first line in memory is top line of RGB-image

	if (hndDriver->hndDec==NULL)
		hndDriver->hndDec = decoder_open();

	if (hndDriver->hndDec)
	{
		configvars_t lpmyconfig[1];
		get_configuration( lpmyconfig );

		decoder_configure( hndDriver->hndDec, color_format, flip_RGB, is_PAL, lpmyconfig );
		hndDriver->DecConfig_id = lpmyconfig->id;

		return ICERR_OK;
	}

	return ICERR_BADFORMAT;
}


LONG my_icm_decompress(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;
	ICDECOMPRESS *icinfo = (ICDECOMPRESS *)lParam1; 
	DWORD dwSize = (DWORD)lParam2;

	if ( (hndDriver==NULL) || (icinfo==NULL) )
		return ICERR_BADPARAM;
	
	if (dwSize<sizeof(ICDECOMPRESS))
		return ICERR_BADPARAM;

	CAutoLockCS LockObj(&hndDriver->csLockDec);

	if (hndDriver->hndDec==NULL)
		return ICERR_BADPARAM;

	if (icinfo->dwFlags & ICDECOMPRESS_NULLFRAME )
		return ICERR_OK;
	
	unsigned char *in = (unsigned char *)icinfo->lpInput;
	unsigned int *out = (unsigned int *)icinfo->lpOutput;

	if ( (in==NULL) || (out==NULL) )
		return ICERR_BADPARAM;

	if ( hndDriver->DecConfig_id != CurrentConfiguration.id )
	{
		if ( my_icm_decompress_begin( hnd, (LPARAM)(icinfo->lpbiInput), (LPARAM)(icinfo->lpbiOutput) ) != ICERR_OK )
			return ICERR_BADFORMAT;
	}

	decode_DVSD( hndDriver->hndDec, in, out );

	return ICERR_OK;
}


LONG my_icm_decompress_end(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * hndDriver = (driver_context_t*)hnd;
	
	if (hndDriver)
	{
		CAutoLockCS LockObj(&hndDriver->csLockDec);

		if (hndDriver->hndDec)
		{
			decoder_close(hndDriver->hndDec);
			hndDriver->hndDec = NULL;
		}
	}

	return ICERR_OK;
}

