/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include <windows.h>

#include "inttypes.h"

#include "config.h"
#include "resource.h"


/****************************************
	global configuration
****************************************/

extern const configvars_t myconfig_default = {
	0,	//id
	QUALITY_MODE_NORMAL,	// encoder_quality_mode
	COLORFORMAT_YUY2 | COLORFORMAT_YV12 | COLORFORMAT_RGB32 | COLORFORMAT_RGB24,	// encoder_input_format;
	0,	// encoder_RGB_to_YUV_conversion = 0;
	YV12_CHROMASAMPLING_MPEG2_I,	// encoder_yv12_chroma_sampling;
	25,	// encoder_dct248_threshold = 10;
	15,	// encoder_dct248_counts = 15;
	0,	// aspect_ratio
	0,	// encoder_coring_enable = 0;
	{	// encoder_coring_matrix[64]
	    0,    0,    0,    0,    0,    0,    0,    0,
	    0,    0,    0,    0,    0,    0,    0,    0,
	    0,    0,    0,    0,    0,    0,    0, 1023,
	    0,    0,    0,    0,    0,    0, 1023, 1023,
	    0,    0,    0,    0,    0, 1023, 1023, 1023,
	    0,    0,    0,    0, 1023, 1023, 1023, 1023,
	    0,    0,    0, 1023, 1023, 1023, 1023, 1023,
	    0,    0, 1023, 1023, 1023, 1023, 1023, 1023
	},
	{  0, 10, 210 },	// encoder_classification_Y[3]
	{  0,  0, 3 },	// encoder_classification_V[3]
	{  0,  0, 3 },	// encoder_classification_U[3]
	COLORFORMAT_YUY2 | COLORFORMAT_YV12 | COLORFORMAT_RGB32 | COLORFORMAT_RGB24,	// decoder_output_format
	0,	// decoder_YUV_to_RGB_conversion = 0;
	YV12_CHROMASAMPLING_MPEG2_I	// decoder_YV12_chroma_sampling;
};

static const configvars_t * lpmyconfig_default = &myconfig_default;


wchar_t *quality_mode_names[NO_OF_QUALITY_MODES] = {
	L"normal", L"fast" };


#define SETTINGS_INI_SECTION L"main"
#define SETTINGS_INI_KEY L"all_021"

int load_configuration( configvars_t *lpmyconfig )
{
	BOOL bReturn;

	memcpy( lpmyconfig, lpmyconfig_default, sizeof(configvars_t) );

	bReturn = GetPrivateProfileStructW(
		SETTINGS_INI_SECTION, 
		SETTINGS_INI_KEY, 
		lpmyconfig, 
		sizeof(configvars_t), 
		CONFIGURATION_FILENAME );

	if ( !(lpmyconfig->encoder_quality_mode<NO_OF_QUALITY_MODES) )
		lpmyconfig->encoder_quality_mode = QUALITY_MODE_NORMAL;

	return 0;
}


int save_configuration( configvars_t *lpmyconfig )
{
	BOOL bReturn;

	bReturn = WritePrivateProfileStructW(
		SETTINGS_INI_SECTION, 
		SETTINGS_INI_KEY, 
		lpmyconfig, 
		sizeof(configvars_t), 
		CONFIGURATION_FILENAME );
	
	return 0;
}


void redraw_dialog(int reset, HWND hwndDlg, const configvars_t *lpmyconfig)
{
	if (reset)
	{
		HWND hctl_quality_mode = GetDlgItem(hwndDlg, IDC_QUALITY_MODE);

		SendMessage(hctl_quality_mode, CB_SETCURSEL, lpmyconfig->encoder_quality_mode, 0);

		// input format
		CheckDlgButton(hwndDlg, IDC_IF_RGB24, (lpmyconfig->encoder_input_format&COLORFORMAT_RGB24) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_IF_RGB32, (lpmyconfig->encoder_input_format&COLORFORMAT_RGB32) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_IF_YUY2, (lpmyconfig->encoder_input_format&COLORFORMAT_YUY2) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_IF_YV12, (lpmyconfig->encoder_input_format&COLORFORMAT_YV12) ? BST_CHECKED : BST_UNCHECKED );
		
		// rgb-to-yuv conversion
		CheckDlgButton(hwndDlg, IDC_RGB2YUV_601, (lpmyconfig->encoder_RGB_to_YUV_conversion==RGB2YUV_601) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_RGB2YUV_UNCHANGED, (lpmyconfig->encoder_RGB_to_YUV_conversion==RGB2YUV_UNCHANGED) ? BST_CHECKED : BST_UNCHECKED );

		// input yv12 chroma sampling
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_IN_DV, (lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_DV) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_PROGRESSIVE, (lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_P) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_INTERLACED, (lpmyconfig->encoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I) ? BST_CHECKED : BST_UNCHECKED );

		// aspect ratio
		CheckDlgButton(hwndDlg, IDC_ASPECTRATIO_4TO3, (lpmyconfig->encoder_aspect_ratio==ASPECTRATIO_4TO3) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_ASPECTRATIO_16TO9, (lpmyconfig->encoder_aspect_ratio!=ASPECTRATIO_4TO3) ? BST_CHECKED : BST_UNCHECKED );

		// output format
		CheckDlgButton(hwndDlg, IDC_OF_RGB24, (lpmyconfig->decoder_output_format&COLORFORMAT_RGB24) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_OF_RGB32, (lpmyconfig->decoder_output_format&COLORFORMAT_RGB32) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_OF_YUY2, (lpmyconfig->decoder_output_format&COLORFORMAT_YUY2) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_OF_YV12, (lpmyconfig->decoder_output_format&COLORFORMAT_YV12) ? BST_CHECKED : BST_UNCHECKED );

		// yuv-to-rgb conversion
		CheckDlgButton(hwndDlg, IDC_YUY2RGB_601, (lpmyconfig->decoder_YUV_to_RGB_conversion==RGB2YUV_601) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_YUY2RGB_UNCHANGED, (lpmyconfig->decoder_YUV_to_RGB_conversion==RGB2YUV_UNCHANGED) ? BST_CHECKED : BST_UNCHECKED );

		// output yv12 chroma sampling
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_OUT_DV, (lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_DV) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_PROGRESSIVE, (lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_P) ? BST_CHECKED : BST_UNCHECKED );
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_INTERLACED, (lpmyconfig->decoder_YV12_chroma_sampling==YV12_CHROMASAMPLING_MPEG2_I) ? BST_CHECKED : BST_UNCHECKED );
	}

	int do_enable_window;
	int is_rgb_input = (IsDlgButtonChecked(hwndDlg, IDC_IF_RGB24) == BST_CHECKED)
		|| (IsDlgButtonChecked(hwndDlg, IDC_IF_RGB32) == BST_CHECKED);
	int is_yv12_input = (IsDlgButtonChecked(hwndDlg, IDC_IF_YV12) == BST_CHECKED);

	do_enable_window = is_rgb_input;
	EnableWindow( GetDlgItem(hwndDlg, IDC_RGB2YUV_601), do_enable_window );
	EnableWindow( GetDlgItem(hwndDlg, IDC_RGB2YUV_UNCHANGED), do_enable_window );

	do_enable_window = is_yv12_input;
	EnableWindow( GetDlgItem(hwndDlg, IDC_YV12_CHROMA_IN_DV), do_enable_window );
	EnableWindow( GetDlgItem(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_PROGRESSIVE), do_enable_window );
	EnableWindow( GetDlgItem(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_INTERLACED), do_enable_window );

	int is_rgb_output = (IsDlgButtonChecked(hwndDlg, IDC_OF_RGB24) == BST_CHECKED)
		|| (IsDlgButtonChecked(hwndDlg, IDC_OF_RGB32) == BST_CHECKED);
	int is_yv12_output = (IsDlgButtonChecked(hwndDlg, IDC_OF_YV12) == BST_CHECKED);

	do_enable_window = is_rgb_output;
	EnableWindow( GetDlgItem(hwndDlg, IDC_YUY2RGB_601), do_enable_window );
	EnableWindow( GetDlgItem(hwndDlg, IDC_YUY2RGB_UNCHANGED), do_enable_window );

	do_enable_window = is_yv12_output;
	EnableWindow( GetDlgItem(hwndDlg, IDC_YV12_CHROMA_OUT_DV), do_enable_window );
	EnableWindow( GetDlgItem(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_PROGRESSIVE), do_enable_window );
	EnableWindow( GetDlgItem(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_INTERLACED), do_enable_window );
}


#define LIMIT_MIN_MAX( x, xmin, xmax ) ( ((x)<xmin) ? xmin : ( ((x)>xmax) ? xmax : (x) ) )

INT_PTR CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	int i;

	configvars_t * lpmyconfig = (configvars_t*)GetWindowLongPtr( hwndDlg, DWLP_USER );
	if (lpmyconfig==NULL) 
	{
		if (uMsg != WM_INITDIALOG)
			return FALSE;

		lpmyconfig = (configvars_t*)lParam;
		SetWindowLongPtr( hwndDlg, DWLP_USER, (LONG_PTR)lpmyconfig );
	}

	if (uMsg == WM_INITDIALOG) 
	{
		// quality mode
		HWND hctl_quality_mode = GetDlgItem(hwndDlg, IDC_QUALITY_MODE);

		if ( SendMessage(hctl_quality_mode, CB_GETCOUNT, 0, 0) == 0 )
		{
			for (i=0; i < NO_OF_QUALITY_MODES; i++) {
			  SendMessageW(hctl_quality_mode, CB_ADDSTRING, 0, (LPARAM)quality_mode_names[i]);
			}
		}

		redraw_dialog(TRUE, hwndDlg, lpmyconfig);
    }
	else if (uMsg == WM_COMMAND) 
	{
		redraw_dialog(FALSE, hwndDlg, lpmyconfig);

		switch (LOWORD(wParam)) {

		case IDOK:

			// quality-mode
			lpmyconfig->encoder_quality_mode = (int)SendMessage(GetDlgItem(hwndDlg, IDC_QUALITY_MODE), CB_GETCURSEL, 0, 0);

			// input format
			lpmyconfig->encoder_input_format = 0;

			if (IsDlgButtonChecked(hwndDlg, IDC_IF_RGB24) == BST_CHECKED)
				lpmyconfig->encoder_input_format |= COLORFORMAT_RGB24;

			if (IsDlgButtonChecked(hwndDlg, IDC_IF_RGB32) == BST_CHECKED)
				lpmyconfig->encoder_input_format |= COLORFORMAT_RGB32;

			if (IsDlgButtonChecked(hwndDlg, IDC_IF_YUY2) == BST_CHECKED)
				lpmyconfig->encoder_input_format |= COLORFORMAT_YUY2;

			if (IsDlgButtonChecked(hwndDlg, IDC_IF_YV12) == BST_CHECKED)
				lpmyconfig->encoder_input_format |= COLORFORMAT_YV12;

			// rgb-to-yuv conversion
			if (IsDlgButtonChecked(hwndDlg, IDC_RGB2YUV_601) == BST_CHECKED)
				lpmyconfig->encoder_RGB_to_YUV_conversion = RGB2YUV_601;
			else 
				lpmyconfig->encoder_RGB_to_YUV_conversion = RGB2YUV_UNCHANGED;

			// input yv12 chroma sampling
			if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_IN_DV) == BST_CHECKED)
				lpmyconfig->encoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_DV;
			else if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_PROGRESSIVE) == BST_CHECKED)
				lpmyconfig->encoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_P;
			else 
				lpmyconfig->encoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_I;

			// aspect ratio
			if (IsDlgButtonChecked(hwndDlg, IDC_ASPECTRATIO_4TO3) == BST_CHECKED)
				lpmyconfig->encoder_aspect_ratio = ASPECTRATIO_4TO3;
			else 
				lpmyconfig->encoder_aspect_ratio = ASPECTRATIO_16TO9;

			// output format
			lpmyconfig->decoder_output_format = 0;

			if (IsDlgButtonChecked(hwndDlg, IDC_OF_RGB24) == BST_CHECKED)
				lpmyconfig->decoder_output_format |= COLORFORMAT_RGB24;
			
			if (IsDlgButtonChecked(hwndDlg, IDC_OF_RGB32) == BST_CHECKED)
				lpmyconfig->decoder_output_format |= COLORFORMAT_RGB32;

			if (IsDlgButtonChecked(hwndDlg, IDC_OF_YUY2) == BST_CHECKED)
				lpmyconfig->decoder_output_format |= COLORFORMAT_YUY2;

			if (IsDlgButtonChecked(hwndDlg, IDC_OF_YV12) == BST_CHECKED)
				lpmyconfig->decoder_output_format |= COLORFORMAT_YV12;

			// yuv-to-rgb conversion
			if (IsDlgButtonChecked(hwndDlg, IDC_YUY2RGB_601) == BST_CHECKED)
				lpmyconfig->decoder_YUV_to_RGB_conversion = RGB2YUV_601;
			else 
				lpmyconfig->decoder_YUV_to_RGB_conversion = RGB2YUV_UNCHANGED;

			// output yv12 chroma sampling
			if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_OUT_DV) == BST_CHECKED)
				lpmyconfig->decoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_DV;
			else if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_PROGRESSIVE) == BST_CHECKED)
				lpmyconfig->decoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_P;
			else 
				lpmyconfig->decoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_I;

			save_configuration( lpmyconfig );

		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			break;

		case IDDEFAULT:
			redraw_dialog(TRUE, hwndDlg, lpmyconfig_default);
			break;

		default:
			break;
		}
	}

	return FALSE;
}


