;******************************************************************************/
;*                                                                            */
;*  Cedocida DV-Codec                                                         */
;*                                                                            */
;*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
;*  Project start: 15.12.2003                                                 */
;*  Last change  : 08.05.2005                                                 */
;*                                                                            */
;******************************************************************************/
;*                                                                            */
;*  This program is free software; you can redistribute it and/or modify      */
;*  it under the terms of the GNU General Public License as published by      */
;*  the Free Software Foundation; either version 2 of the License, or         */
;*  (at your option) any later version.                                       */
;*                                                                            */
;*  This program is distributed in the hope that it will be useful,           */
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
;*  GNU General Public License for more details.                              */
;*                                                                            */
;*  You should have received a copy of the GNU General Public License         */
;*  along with this program; if not, write to the Free Software               */
;*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
;*                                                                            */
;******************************************************************************/

%include "nasm.inc"


;=============================================================================
; Read only data
;=============================================================================

SECTION .rodata data align=16

mask1: 
	dw 0x00ff, 0x00ff, 0x00ff, 0x00ff	;only the first 8 byte are used but due to alignment we double it
	dw 0x00ff, 0x00ff, 0x00ff, 0x00ff

mask2: 
	dw 0xff00, 0xff00, 0xff00, 0xff00
	dw 0xff00, 0xff00, 0xff00, 0xff00

add_2:
	dw 2, 2, 2 ,2
	dw 2, 2, 2 ,2

add_1:
	dw 1, 1, 1 ,1
	dw 1, 1, 1 ,1



;=============================================================================
; Code
;=============================================================================

SECTION .text

;void mmx_yv12_to_yuy2(
;	const uint8_t* srcY, 
;	const uint8_t* srcU, 
;	const uint8_t* srcV, 
;	int src_rowsize, 
;	int src_pitch, 
;	int src_pitch_uv, 
;	uint8_t* dst, 
;	int dst_pitch,
;	int height)

%macro YV12_TO_YUY2__HEADER 1
		%define NUM_PUSH %1
		%define NUM_ARGS 9
		%define NUM_VARS 13

		%ifdef WIN64	
		%define NBYTES 8
		mov [rsp + (NUM_PUSH + 1)*8], rcx
		mov [rsp + (NUM_PUSH + 2)*8], rdx
		mov [rsp + (NUM_PUSH + 3)*8], r8
		mov [rsp + (NUM_PUSH + 4)*8], r9
		%else
		%define NBYTES 4
		%endif
	 
		sub rsp, NUM_VARS*NBYTES
		
		%define srcY         (rsp + (NUM_PUSH + NUM_VARS + 1)*NBYTES)
		%define srcU         (rsp + (NUM_PUSH + NUM_VARS + 2)*NBYTES)
		%define srcV         (rsp + (NUM_PUSH + NUM_VARS + 3)*NBYTES)
		%define src_rowsize  (rsp + (NUM_PUSH + NUM_VARS + 4)*NBYTES)
		%define src_pitch    (rsp + (NUM_PUSH + NUM_VARS + 5)*NBYTES)
		%define src_pitch_uv (rsp + (NUM_PUSH + NUM_VARS + 6)*NBYTES)
		%define dst          (rsp + (NUM_PUSH + NUM_VARS + 7)*NBYTES)
		%define dst_pitch    (rsp + (NUM_PUSH + NUM_VARS + 8)*NBYTES)
		%define height       (rsp + (NUM_PUSH + NUM_VARS + 9)*NBYTES)

		%define srcp          (rsp + 0*NBYTES)
		%define src_pitch2    (rsp + 4*NBYTES)
		%define src_pitch4    (rsp + 5*NBYTES)
		%define src_pitch_uv2 (rsp + 6*NBYTES)
		%define src_pitch_uv4 (rsp + 7*NBYTES)
		%define dst_pitch2    (rsp + 8*NBYTES)
		%define dst_pitch4    (rsp + 9*NBYTES)
		%define skipnext      (rsp +10*NBYTES)
		%define x             (rsp +11*NBYTES)
		%define y             (rsp +12*NBYTES)

		;srcp[0]=srcY;
		mov rax, [srcY]
		mov [srcp], rax
		
		;srcp[1]=srcU;
		mov rax, [srcU]
		mov [srcp+1*NBYTES], rax

		;srcp[2]=srcV;
		mov rax, [srcV]
		mov [srcp+2*NBYTES], rax

		movsxd rax, DWORD [src_rowsize]
		mov [src_rowsize], rax

		movsxd rax, DWORD [src_pitch]
		mov [src_pitch], rax
		shl rax, 1
		mov [src_pitch2], rax
		shl rax, 1
		mov [src_pitch4], rax

		movsxd rax, DWORD [src_pitch_uv]
		mov [src_pitch_uv], rax
		shl rax, 1
		mov [src_pitch_uv2], rax
		shl rax, 1
		mov [src_pitch_uv4], rax

		movsxd rax, DWORD [dst_pitch]
		mov [dst_pitch], rax
		shl rax, 1
		mov [dst_pitch2], rax
		shl rax, 1
		mov [dst_pitch4], rax

		movsxd rax, DWORD [height]
		mov [height], rax

		;int x=0;
		;int y=0;
		;int skipnext=0;
		xor rax, rax
		mov [x], rax
		mov [y], rax
		mov [skipnext], rax
%endmacro	


;*************************************
;* Progressive YV12 -> YUY2 conversion
;*
;* (c) 2003, Klaus Post.
;*
;* Requires mod 8 rowsize.
;* MMX version.
;*************************************/
;
;void mmx_yv12_to_yuy2(
;	const uint8_t* srcY, 
;	const uint8_t* srcU, 
;	const uint8_t* srcV, 
;	int src_rowsize, 
;	int src_pitch, 
;	int src_pitch_uv, 
;	uint8_t* dst, 
;	int dst_pitch,
;	int height)


cfunction mmx_yv12_to_yuy2
		push rsi
		push rdi
		push rbx
	
		YV12_TO_YUY2__HEADER 3

		;line 0
		mov rdi, [dst]
		mov rax, [srcY]
		mov rbx, [srcU]
		mov rcx, [srcV]
		call mmx_yv12_to_yuy2_line
		
		;line 1
		mov rdi, [dst_pitch]
		add rdi, [dst]
		mov rax, [src_pitch]	
		add rax, [srcY]
		mov rbx, [srcU]
		mov rcx, [srcV]
		call mmx_yv12_to_yuy2_line
		
		;/****************************************
		;* Conversion main loop.
		;* The code properly interpolates UV from
		;* interlaced material.
		;* We process two lines in the same field
		;* in the same loop, to avoid reloading
		;* chroma each time.
		;*****************************************/

		mov rax, [height]
		sub rax, 4
		mov [height], rax

		lea rsi, [srcp]
		mov rdi, [dst]
		add rdi, [dst_pitch2]
		mov [dst], rdi

		mov rax, [rsi+0*NBYTES]
		add rax, [src_pitch2]
		mov [rsi+0*NBYTES], rax
		
		mov rbx, [rsi+1*NBYTES]
		mov rcx, [rsi+2*NBYTES]
		
		mov rdx,0
		jmp .yloop_test
		
		align 16
	.yloop:
		mov rdx, 0               ; x counter
		jmp .xloop_test
		
		align 16
	.xloop:
		movsxd rdx, DWORD [src_pitch_uv]
		movq mm0,[rax]          ; mm0 = Y current line
		pxor mm7,mm7
		movd mm2,[rbx+rdx]            ; mm2 = U top field
		movd mm3, [rcx+rdx]          ; mm3 = V top field
		movd mm4,[rbx]        ; U prev top field
		movq mm1,mm0             ; mm1 = Y current line
		movd mm5,[rcx]        ; V prev top field

		punpcklbw mm2,mm7        ; U 00uu 00uu 00uu 00uu
		punpcklbw mm3,mm7         ; V 00vv 00vv 00vv 00vv
		punpcklbw mm4,mm7        ; U 00uu 00uu 00uu 00uu
		punpcklbw mm5,mm7         ; V 00vv 00vv 00vv 00vv
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4, [add_2]
		paddusw mm5, [add_2]
		psrlw mm4,2
		psrlw mm5,2

		punpcklbw mm0,mm7        ; Y low
		punpckhbw mm1,mm7         ; Y high*
		pxor mm6,mm6
		punpcklbw mm6,mm4         ; U 0000 uu00 0000 uu00 (low)
		punpckhbw mm7,mm4         ; V 0000 uu00 0000 uu00 (high
		por mm0,mm6
		por mm1,mm7
		movq mm6,mm5
		punpcklbw mm5,mm5          ; V 0000 vvvv 0000 vvvv (low)
		punpckhbw mm6,mm6           ; V 0000 vvvv 0000 vvvv (high)
		pslld mm5,24
		pslld mm6,24
		por mm0,mm5
		por mm1,mm6
		mov rdx, [src_pitch_uv2]
		movq [rdi],mm0
		movq [rdi+8],mm1

		;Next line 
		movd mm4,[rbx+rdx]        ; U next top field
		movd mm5,[rcx+rdx]       ; V prev top field
		mov rdx, [src_pitch]
		pxor mm7,mm7
		movq mm0,[rax+rdx]        ; Next U-line
		movq mm1,mm0             ; mm1 = Y current line

		punpcklbw mm4,mm7        ; U 00uu 00uu 00uu 00uu
		punpcklbw mm5,mm7         ; V 00vv 00vv 00vv 00vv
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4, [add_2]
		paddusw mm5, [add_2]
		psrlw mm4,2
		psrlw mm5,2

		punpcklbw mm0,mm7        ; Y low
		punpckhbw mm1,mm7         ; Y high*
		pxor mm6,mm6
		punpcklbw mm6,mm4         ; U 0000 uu00 0000 uu00 (low)
		punpckhbw mm7,mm4         ; V 0000 uu00 0000 uu00 (high
		por mm0,mm6
		por mm1,mm7
		movq mm6,mm5
		punpcklbw mm5,mm5          ; V 0000 vvvv 0000 vvvv (low)
		punpckhbw mm6,mm6           ; V 0000 vvvv 0000 vvvv (high)
		pslld mm5,24
		mov rdx, [dst_pitch]
		pslld mm6,24
		por mm0,mm5
		por mm1,mm6
		movq [rdi+rdx],mm0
		movq [rdi+rdx+8],mm1
		add rdi, BYTE 16
		mov rdx, [x]
		add rax, BYTE 8
		add rbx, BYTE 4
		add rdx, BYTE 8
		add rcx, BYTE 4
	
	.xloop_test:
		cmp edx, [src_rowsize]
		mov [x], rdx
		jl .xloop
		
		mov rdi, [dst]
		mov rax,[rsi+0*NBYTES]

		add rdi, [dst_pitch2]
		add rax, [src_pitch2]
		
		mov rbx, [src_pitch_uv]
		mov rcx, [src_pitch_uv]
		add rbx, [rsi+1*NBYTES]
		add rcx, [rsi+2*NBYTES]
		
		mov rdx, [y]
		mov [rsi+0*NBYTES],rax
		mov [rsi+1*NBYTES],rbx
		mov [rsi+2*NBYTES],rcx
		mov [dst],rdi
		add rdx, BYTE 2

	.yloop_test:
		cmp edx, [height]
		mov [y], rdx
		jl .yloop

		;line 0
		mov rdi, [dst]
		mov rax, [rsi+0*NBYTES]
		mov rbx, [src_pitch_uv]
		mov rcx, [src_pitch_uv]
		add rbx, [rsi+1*NBYTES]
		add rcx, [rsi+2*NBYTES]
		call mmx_yv12_to_yuy2_line
		
		;line 1

		mov rdi, [dst_pitch]
		add rdi, [dst]
		mov rax, [src_pitch]
		add rax, [rsi+0*NBYTES]
		mov rbx, [src_pitch_uv]
		mov rcx, [src_pitch_uv]
		add rbx, [rsi+1*NBYTES]
		add rcx, [rsi+2*NBYTES]
		call mmx_yv12_to_yuy2_line

		add rsp, NUM_VARS*NBYTES		
		pop rbx
		pop rdi
		pop rsi
		ret

mmx_yv12_to_yuy2_line:
		xor rdx, rdx
		pxor mm7,mm7
		jmp .xloop_test_p
		
	.xloop_p:
		movq mm0,[rax]    ;Y
		movd mm1,[rbx]  ;U
		movq mm3,mm0  
		movd mm2,[rcx]   ;V
		punpcklbw mm0,mm7  ; Y low
		punpckhbw mm3,mm7   ; Y high
		punpcklbw mm1,mm7   ; 00uu 00uu
		punpcklbw mm2,mm7   ; 00vv 00vv
		movq mm4,mm1
		movq mm5,mm2
		punpcklbw mm1,mm7   ; 0000 00uu low
		punpcklbw mm2,mm7   ; 0000 00vv low
		punpckhbw mm4,mm7   ; 0000 00uu high
		punpckhbw mm5,mm7   ; 0000 00vv high
		pslld mm1,8
		pslld mm4,8
		pslld mm2,24
		pslld mm5,24
		por mm0, mm1
		por mm3, mm4
		por mm0, mm2
		por mm3, mm5
		movq [rdi],mm0
		movq [rdi+8],mm3
		add rax, BYTE 8
		add rbx, BYTE 4
		add rcx, BYTE 4
		add rdx, BYTE 8
		add rdi, BYTE 16
		
	.xloop_test_p:
		cmp edx, DWORD [src_rowsize+1*NBYTES]
		jl .xloop_p			
		ret


;*************************************
;* Interlaced YV12 -> YUY2 conversion
;*
;* (c) 2003, Klaus Post.
;*
;* Requires mod 8 rowsize.
;* MMX version.
;*************************************
;
;void mmx_yv12_i_to_yuy2(
;	const uint8_t* srcY, 
;	const uint8_t* srcU, 
;	const uint8_t* srcV, 
;	int src_rowsize, 
;	int src_pitch, 
;	int src_pitch_uv,     
;	uint8_t* dst,
;	int dst_pitch,                 
;	int height) 

cfunction mmx_yv12_i_to_yuy2
		push rsi
		push rdi
		push rbx
	
		YV12_TO_YUY2__HEADER 3

		; **** Do first and last lines - NO interpolation:   *****

		mov rdi, [dst]
		mov rax, [srcY]
		mov rbx, [srcU]
		mov rcx, [srcV]
		call mmx_yv12_i_to_yuy2_line
		
		add rdi, [dst_pitch2]
		add rax, [src_pitch2]
		call mmx_yv12_i_to_yuy2_line
		
		sub rdi, [dst_pitch]
		sub rax, [src_pitch]
		add rbx, [src_pitch_uv]
		add rcx, [src_pitch_uv]
		call mmx_yv12_i_to_yuy2_line

		add rdi, [dst_pitch2]
		add rax, [src_pitch2]
		call mmx_yv12_i_to_yuy2_line
	
		mov rdi, [height]
		sub rdi, 4	
		imul rdi, [dst_pitch]
		add rdi, [dst]
	
		mov rax, [height]
		sub rax, 4	
		imul rax, [src_pitch]
		add rax, [srcY]
		
		mov rbx, [height]
		sar rbx, 1
		sub rbx, 2	
		imul rbx, [src_pitch_uv]
		add rbx, [srcU]
		
		mov rcx, [height]
		sar rcx, 1
		sub rcx, 2	
		imul rcx, [src_pitch_uv]
		add rcx, [srcV]
	
		call mmx_yv12_i_to_yuy2_line
		
		add rdi, [dst_pitch2]
		add rax, [src_pitch2]
		call mmx_yv12_i_to_yuy2_line
		
		sub rdi, [dst_pitch]
		sub rax, [src_pitch]
		add rbx, [src_pitch_uv]
		add rcx, [src_pitch_uv]
		call mmx_yv12_i_to_yuy2_line

		add rdi, [dst_pitch2]
		add rax, [src_pitch2]
		call mmx_yv12_i_to_yuy2_line

		;****************************************
		;* Conversion main loop.
		;* The code properly interpolates UV from
		;* interlaced material.
		;* We process two lines in the same field
		;* in the same loop, to avoid reloading
		;* chroma each time.
		;*****************************************
		
		mov rax, [height]
		sub rax, 8
		mov [height], rax

		mov rax, [dst_pitch4]
		add [dst], rax
		mov rax, [src_pitch4]
		add [srcY], rax
		mov rax, [src_pitch_uv2]
		add [srcU], rax
		add [srcV], rax
		
		mov rax, [srcY]
		mov [srcp + 0*NBYTES], rax
		mov rax, [srcU]
		sub rax, [src_pitch_uv2]
		mov [srcp + 1*NBYTES], rax
		mov rax, [srcV]
		sub rax, [src_pitch_uv2]
		mov [srcp + 2*NBYTES], rax

		lea rsi, [srcp]
		mov rdi, [dst]

		mov rax, [rsi+0*NBYTES]
		mov rbx, [rsi+1*NBYTES]
		mov rcx, [rsi+2*NBYTES]
		xor rdx, rdx
		jmp .yloop_test
		
		align 16
	.yloop:
		xor rdx, rdx               ; x counter
		jmp .xloop_test
		
		align 16
	.xloop:
		mov rdx, [src_pitch_uv2]
		movq mm0, [rax]          ; mm0 = Y current line
		pxor mm7,mm7
		movd mm2, [rbx+rdx]            ; mm2 = U top field
		movd mm3, [rcx+rdx]          ; mm3 = V top field
		movd mm4, [rbx]        ; U prev top field
		movq mm1,mm0             ; mm1 = Y current line
		movd mm5,[rcx]        ; V prev top field

		punpcklbw mm2,mm7        ; U 00uu 00uu 00uu 00uu
		punpcklbw mm3,mm7         ; V 00vv 00vv 00vv 00vv
		punpcklbw mm4,mm7        ; U 00uu 00uu 00uu 00uu
		punpcklbw mm5,mm7         ; V 00vv 00vv 00vv 00vv
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4, [add_2]
		paddusw mm5, [add_2]
		psrlw mm4,2
		pxor mm7,mm7
		psrlw mm5,2

		punpcklbw mm0,mm7        ; Y low
		punpckhbw mm1,mm7         ; Y high*
		pxor mm6,mm6
		punpcklbw mm6,mm4         ; U 0000 uu00 0000 uu00 (low)
		punpckhbw mm7,mm4         ; V 0000 uu00 0000 uu00 (high
		por mm0,mm6
		por mm1,mm7
		movq mm6,mm5
		punpcklbw mm5,mm5          ; V 0000 vvvv 0000 vvvv (low)
		punpckhbw mm6,mm6           ; V 0000 vvvv 0000 vvvv (high)
		pslld mm5,24
		pslld mm6,24
		por mm0, mm5
		por mm1, mm6
		mov rdx, [src_pitch_uv4]
		movq [rdi+0], mm0
		movq [rdi+8], mm1

		;Next line in same field
		movd mm4, [rbx+rdx]        ; U next top field
		movd mm5, [rcx+rdx]       ; V prev top field
		mov rdx, [src_pitch2]
		pxor mm7,mm7
		movq mm0, [rax+rdx]        ; Next U-line
		movq mm1,mm0             ; mm1 = Y current line

		punpcklbw mm4,mm7        ; U 00uu 00uu 00uu 00uu
		punpcklbw mm5,mm7         ; V 00vv 00vv 00vv 00vv
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4,mm2
		paddusw mm5,mm3
		paddusw mm4, [add_2]
		paddusw mm5, [add_2]
		psrlw mm4,2
		pxor mm7,mm7
		psrlw mm5,2

		punpcklbw mm0,mm7        ; Y low
		punpckhbw mm1,mm7         ; Y high*
		pxor mm6,mm6
		punpcklbw mm6,mm4         ; U 0000 uu00 0000 uu00 (low)
		punpckhbw mm7,mm4         ; V 0000 uu00 0000 uu00 (high
		por mm0,mm6
		por mm1,mm7
		movq mm6,mm5
		punpcklbw mm5,mm5          ; V 0000 vvvv 0000 vvvv (low)
		punpckhbw mm6,mm6           ; V 0000 vvvv 0000 vvvv (high)
		pslld mm5,24
		mov rdx,[dst_pitch2]
		pslld mm6,24
		por mm0,mm5
		por mm1,mm6
		movq [rdi+rdx],mm0
		movq [rdi+rdx+8],mm1
		add rdi,16
		mov rdx, [x]
		add rax, BYTE 8
		add rbx, BYTE 4
		add rdx, BYTE 8
		add rcx, BYTE 4
		
	.xloop_test:
		cmp rdx, [src_rowsize]
		mov [x], rdx
		jl .xloop
		
		mov rdi, [dst]
		mov rax,[rsi+0*NBYTES]
		mov rbx,[rsi+1*NBYTES]
		mov rcx,[rsi+2*NBYTES]

		mov rdx, [skipnext]
		cmp rdx, 1
		je .dont_skip
		
		add rdi,[dst_pitch]
		add rax,[src_pitch]
		add rbx,[src_pitch_uv]
		add rcx,[src_pitch_uv]
		mov [skipnext], BYTE 1
		jmp .yloop  ; Never out of loop, if not skip
		
		align 16
	.dont_skip:
		add rdi,[dst_pitch4]
		add rax,[src_pitch4]
		add rbx,[src_pitch_uv2]
		add rcx,[src_pitch_uv2]
		mov [skipnext], BYTE 0
		mov rdx, [y]
		mov [rsi+0*NBYTES],rax
		mov [rsi+1*NBYTES],rbx
		mov [rsi+2*NBYTES],rcx
		mov [dst],rdi
		add rdx, BYTE 4

	.yloop_test:
		cmp rdx, [height]
		mov [y], rdx
		jl .yloop

		add rsp, NUM_VARS*NBYTES		
		pop rbx
		pop rdi
		pop rsi
		ret

mmx_yv12_i_to_yuy2_line:
		push rdi
		push rax
		push rbx
		push rcx

		xor rdx, rdx
		pxor mm7, mm7
		jmp .xloop_test_p
		
	.xloop_p:
		movq mm0,[rax]    ;Y
		movd mm1,[rbx]  ;U
		movq mm3,mm0  
		movd mm2,[rcx]   ;V
		punpcklbw mm0,mm7  ; Y low
		punpckhbw mm3,mm7   ; Y high
		punpcklbw mm1,mm7   ; 00uu 00uu
		punpcklbw mm2,mm7   ; 00vv 00vv
		movq mm4,mm1
		movq mm5,mm2
		punpcklbw mm1,mm7   ; 0000 00uu low
		punpcklbw mm2,mm7   ; 0000 00vv low
		punpckhbw mm4,mm7   ; 0000 00uu high
		punpckhbw mm5,mm7   ; 0000 00vv high
		pslld mm1,8
		pslld mm4,8
		pslld mm2,24
		pslld mm5,24
		por mm0, mm1
		por mm3, mm4
		por mm0, mm2
		por mm3, mm5
		movq [rdi],mm0
		movq [rdi+8],mm3
		add rax, BYTE 8
		add rbx, BYTE 4
		add rcx, BYTE 4
		add rdx, BYTE 8
		add rdi, BYTE 16
		
	.xloop_test_p:
		cmp rdx, [src_rowsize+5*NBYTES]
		jl .xloop_p
		
		pop rcx
		pop rbx
		pop rax
		pop rdi
		ret



%macro YUY2_TO_YV12__HEADER 1
		%define NUM_PUSH %1
		%define NUM_ARGS 9
		%define NUM_VARS 10

		%ifdef WIN64	
		%define NBYTES 8
		mov [rsp + (NUM_PUSH + 1)*8], rcx
		mov [rsp + (NUM_PUSH + 2)*8], rdx
		mov [rsp + (NUM_PUSH + 3)*8], r8
		mov [rsp + (NUM_PUSH + 4)*8], r9
		%else
		%define NBYTES 4
		%endif
	 
		sub rsp, NUM_VARS*NBYTES
		
		%define src         (rsp + (NUM_PUSH + NUM_VARS + 1)*NBYTES)
		%define src_rowsize (rsp + (NUM_PUSH + NUM_VARS + 2)*NBYTES)
		%define src_pitch   (rsp + (NUM_PUSH + NUM_VARS + 3)*NBYTES)
		%define dstY        (rsp + (NUM_PUSH + NUM_VARS + 4)*NBYTES)
		%define dstU        (rsp + (NUM_PUSH + NUM_VARS + 5)*NBYTES)
		%define dstV        (rsp + (NUM_PUSH + NUM_VARS + 6)*NBYTES)
		%define dst_pitchY  (rsp + (NUM_PUSH + NUM_VARS + 7)*NBYTES)
		%define dst_pitchUV (rsp + (NUM_PUSH + NUM_VARS + 8)*NBYTES)
		%define height      (rsp + (NUM_PUSH + NUM_VARS + 9)*NBYTES)

		%define dstp        (rsp + 0*NBYTES)
		%define src_pitch2  (rsp + 4*NBYTES)
		%define src_pitch4  (rsp + 5*NBYTES)
		%define dst_pitch2  (rsp + 6*NBYTES)
		%define dst_pitch3  (rsp + 7*NBYTES)
		%define y           (rsp + 8*NBYTES)

		;dstp[0]=dstY;
		mov rax, [dstY]
		mov [dstp], rax
		
		;dstp[2]=dstU;
		mov rax, [dstU]
		mov [dstp+2*NBYTES], rax

		;dstp[3]=dstV;
		mov rax, [dstV]
		mov [dstp+3*NBYTES], rax

		movsxd rax, DWORD [src_pitch]
		mov [src_pitch], rax
		shl rax, 1
		mov [src_pitch2], rax
		shl rax, 1
		mov [src_pitch4], rax
		
		movsxd rax, DWORD [dst_pitchY]
		mov [dst_pitchY], rax
		shl rax, 1
		mov [dst_pitch2], rax

		movsxd rax, DWORD [dst_pitchUV]
		mov [dst_pitchUV], rax

		movsxd rax, DWORD [height]
		mov [height], rax

		mov rax, [dst_pitchY]
		imul rax, 3
		mov [dst_pitch3], rax

		;int y=0;
		xor rax, rax
		mov [y], rax
		
		;src_rowsize = (src_rowsize+3)/4;
		movsxd rax, DWORD [src_rowsize]
		sar rax, 2
		mov [src_rowsize], rax
%endmacro	


;/********************************
;* Progressive YUY2 to YV12
;* 
;* (c) Copyright 2003, Klaus Post
;*
;* Converts 8x2 (8 pixels, two lines) in parallel.
;* Requires mod8 pitch for output, and mod16 pitch for input.
;* MMX Version (much slower than ISSE!) (used as fallback for ISSE version)
;********************************/

;void mmx_yuy2_to_yv12(
;	const uint8_t* src, 
;	int src_rowsize, 
;	int src_pitch, 
;	uint8_t* dstY, 
;	uint8_t* dstU, 
;	uint8_t* dstV, 
;	int dst_pitchY, 
;	int dst_pitchUV,
;	int height) 

cfunction mmx_yuy2_to_yv12
		push rsi
		push rdi
		push rbx

		YUY2_TO_YV12__HEADER 3

		;dstp[1]=dstY+dst_pitchY;
		mov rax, [dstY]
		add rax, [dst_pitchY]
		mov [dstp+1*NBYTES], rax

		mov rdx, 0
		mov rsi, [src]
		lea rdi, [dstp]

		jmp .yloop_test
		
		align 16
	.yloop:
		mov rdx,0               ; x counter   
		mov rax, [src_pitch]
		jmp .xloop_test
		
		align 16
	.xloop:      
		movq mm0,[rsi]        ; YUY2 upper line  (4 pixels luma, 2 chroma)
		movq mm1,[rsi+rax]   ; YUY2 lower line  
		movq mm6,mm0
		movq mm2, [rsi+8]    ; Load second pair
		movq mm3, [rsi+rax+8]
		movq mm5,mm2
		movq mm7, mm1
		movq mm4, mm3
		psrlw mm5,8
		psrlw mm6,8
		psrlw mm4,8
		psrlw mm7,8
		paddw mm5,mm4
		paddw mm6,mm7
		movq mm4,[mask1]
		movq mm7,[mask2]
		paddw mm5,[add_1]
		paddw mm6,[add_1]

		pand mm0,mm4          ; Mask luma
		psrlw mm5,1
		pand mm1,mm4          ; Mask luma
		psrlw mm6,1
		pand mm2,mm4          ; Mask luma
		pand mm3,mm4         
		packuswb mm0, mm2     ; Pack luma (upper)
		packuswb mm6, mm5    ; Pack chroma
		packuswb mm1, mm3     ; Pack luma (lower)     
		movq mm5, mm6        ; Chroma copy
		pand mm5, mm7         ; Mask V
		pand mm6, mm4        ; Mask U
		psrlq mm5,8            ; shift down V
		packuswb mm5, mm7     ; Pack U 
		packuswb mm6, mm7     ; Pack V 
		mov rbx, [rdi+0*NBYTES]
		mov rcx, [rdi+1*NBYTES]
		movq [rbx+rdx*2],mm0
		movq [rcx+rdx*2],mm1

		mov rcx, [rdi+2*NBYTES]
		mov rbx, [rdi+3*NBYTES]
		movd [rcx+rdx], mm6  ; Store U
		movd [rbx+rdx], mm5   ; Store V
		add rsi, BYTE 16
		add rdx, BYTE 4

	.xloop_test:
		cmp rdx,[src_rowsize]
		jl .xloop

		mov rsi, [src]
		mov rdx,[rdi+0*NBYTES]
		mov rbx,[rdi+1*NBYTES]
		mov rcx,[rdi+2*NBYTES]
		mov rax,[rdi+3*NBYTES]

		add rdx, [dst_pitch2]
		add rbx, [dst_pitch2]
		add rcx, [dst_pitchUV]
		add rax, [dst_pitchUV]
		add rsi, [src_pitch2]

		mov [rdi+0*NBYTES],rdx
		mov [rdi+1*NBYTES],rbx
		mov [rdi+2*NBYTES],rcx
		mov [rdi+3*NBYTES],rax
		mov rdx, [y]
		mov [src], rsi

		add rdx, BYTE 2

	.yloop_test:
		cmp rdx,[height]
		mov [y],rdx
		jl .yloop
		
		add rsp, NUM_VARS*NBYTES		
		pop rbx
		pop rdi
		pop rsi
		ret


;/********************************
;* Interlaced YUY2 to YV12
;* 
;* (c) Copyright 2003, Klaus Post
;*
;* Converts 8x2 (8 pixels, two lines) in parallel.
;* Requires mod8 pitch for output, and mod16 pitch for input.
;* MMX version (used as fallback for ISSE version)
;********************************/

;void mmx_yuy2_i_to_yv12(
;	const uint8_t* src, 
;	int src_rowsize, 
;	int src_pitch, 
;	uint8_t* dstY, 
;	uint8_t* dstU, 
;	uint8_t* dstV, 
;	int dst_pitchY, 
;	int dst_pitchUV,
;	int height)

cfunction mmx_yuy2_i_to_yv12
		push rsi
		push rdi
		push rbx

		YUY2_TO_YV12__HEADER 3

		;dstp[1]=dstY+(dst_pitchY*2);
		mov rax, [dstY]
		add rax, [dst_pitch2]
		mov [dstp+1*NBYTES], rax
		
		mov rdx, 0
		mov rsi, [src]
		lea rdi, [dstp]
		
		jmp .yloop_test
		
		align 16		
	.yloop:
		mov rdx,0               ; x counter   
		mov rax, [src_pitch2]
		
		jmp .xloop_test
		align 16

	.xloop:      
		movq mm0,[rsi]        ; YUY2 upper line  (4 pixels luma, 2 chroma) (u1)
		movq mm1,[rsi+rax]   ; YUY2 lower line  (l1)
		movq mm7,mm0          ; (u1)
		movq mm2, [rsi+8]    ; Load second pair (u2)
		movq mm3, [rsi+rax+8] ; (l2)
		movq mm4,mm2         ; (u2)
		movq mm6, mm1         ; (l1)
		movq mm5, mm3        ; (l2)

		psrlw mm5,8  ;(l2)
		psrlw mm6,8 ;(l1)
		psrlw mm4,8  ;(u2)
		psrlw mm7,8 ;(u1)

		paddw mm5,mm4  ;l2+u2
		paddw mm6,mm7 ;l1+u1
		paddw mm5,mm4  ;l2+u2+u2
		paddw mm6,mm7 ;l1+u1+u1
		paddw mm5,mm4  ;l2+u2+u2+u2
		paddw mm6,mm7 ;l1+u1+u1+u1
		movq mm4,[mask1]
		movq mm7,[mask2]
		paddw mm5,[add_2]
		paddw mm6,[add_2]

		pand mm0,mm4          ; Mask luma
		psrlw mm5, 2
		pand mm1,mm4          ; Mask luma
		psrlw mm6, 2
		pand mm2,mm4          ; Mask luma
		pand mm3,mm4         
		pand mm5,mm4           ; Mask chroma
		pand mm6,mm4          ; Mask chroma
		packuswb mm0, mm2     ; Pack luma (upper)
		packuswb mm6, mm5    ; Pack chroma
		packuswb mm1, mm3     ; Pack luma (lower)     
		movq mm5, mm6        ; Chroma copy
		pand mm5, mm7         ; Mask V
		pand mm6, mm4        ; Mask U
		psrlq mm5,8            ; shift down V
		packuswb mm5, mm7     ; Pack U 
		packuswb mm6, mm7     ; Pack V 
		mov rbx, [rdi+0*NBYTES]
		mov rcx, [rdi+1*NBYTES]
		movq [rbx+rdx*2],mm0
		movq [rcx+rdx*2],mm1

		mov rcx, [rdi+2*NBYTES]
		mov rbx, [rdi+3*NBYTES]
		movd [rcx+rdx], mm6  ; Store U
		movd [rbx+rdx], mm5   ; Store V
		add rsi, BYTE 16
		add rdx, BYTE 4

	.xloop_test:
		cmp rdx,[src_rowsize]
		jl .xloop

		mov rsi, [src]
		mov rdx,[rdi+0*NBYTES]
		mov rbx,[rdi+1*NBYTES]
		mov rcx,[rdi+2*NBYTES]
		mov rax,[rdi+3*NBYTES]

		add rdx, [dst_pitchY]
		add rbx, [dst_pitchY]
		add rcx, [dst_pitchUV]
		add rax, [dst_pitchUV]
		add rsi, [src_pitch]

		mov [rdi+0*NBYTES],rdx
		mov [rdi+1*NBYTES],rbx
		mov [rdi+2*NBYTES],rcx
		mov [rdi+3*NBYTES],rax

		mov rdx, 0
		mov rax, [src_pitch2]

		jmp .xloop2_test

	.xloop2:   ; Second field
		movq mm0,[rsi]        ; YUY2 upper line  (4 pixels luma, 2 chroma)
		movq mm1,[rsi+rax]   ; YUY2 lower line  
		movq mm6,mm0
		movq mm2, [rsi+8]    ; Load second pair
		movq mm3, [rsi+rax+8]
		movq mm5,mm2
		movq mm7, mm1
		movq mm4, mm3
		psrlw mm5,8
		psrlw mm6,8
		psrlw mm4,8
		psrlw mm7,8
		paddw mm5,mm4
		paddw mm6,mm7
		paddw mm5,mm4
		paddw mm6,mm7
		paddw mm5,mm4
		paddw mm6,mm7
		movq mm4,[mask1]
		movq mm7,[mask2]
		paddw mm5,[add_2]
		paddw mm6,[add_2]

		pand mm0,mm4          ; Mask luma
		psrlw mm5, 2
		pand mm1,mm4          ; Mask luma
		psrlw mm6, 2
		pand mm2,mm4          ; Mask luma
		pand mm3,mm4         
		pand mm5,mm4           ; Mask chroma
		pand mm6,mm4          ; Mask chroma
		packuswb mm0, mm2     ; Pack luma (upper)
		packuswb mm6, mm5    ; Pack chroma
		packuswb mm1, mm3     ; Pack luma (lower)     
		movq mm5, mm6        ; Chroma copy
		pand mm5, mm7         ; Mask V
		pand mm6, mm4        ; Mask U
		psrlq mm5,8            ; shift down V
		packuswb mm5, mm7     ; Pack U 
		packuswb mm6, mm7     ; Pack V 
		mov rbx, [rdi+0*NBYTES]
		mov rcx, [rdi+1*NBYTES]
		movq [rbx+rdx*2],mm0
		movq [rcx+rdx*2],mm1

		mov rcx, [rdi+2*NBYTES]
		mov rbx, [rdi+3*NBYTES]
		movd [rcx+rdx], mm6  ; Store U
		movd [rbx+rdx], mm5   ; Store V
		add rsi, BYTE 16
		add rdx, BYTE 4

	.xloop2_test:
		cmp rdx,[src_rowsize]
		jl .xloop2

		mov rsi, [src]
		mov rdx,[rdi+0*NBYTES]
		mov rbx,[rdi+1*NBYTES]
		mov rcx,[rdi+2*NBYTES]
		mov rax,[rdi+3*NBYTES]

		add rdx, [dst_pitch3]
		add rbx, [dst_pitch3]
		add rcx, [dst_pitchUV]
		add rax, [dst_pitchUV]
		add rsi, [src_pitch4]

		mov [rdi+0*NBYTES],rdx
		mov [rdi+1*NBYTES],rbx
		mov [rdi+2*NBYTES],rcx
		mov [rdi+3*NBYTES],rax
		mov rdx, [y]
		mov [src],rsi

		add rdx, BYTE 4

	.yloop_test:
		cmp rdx, [height]
		mov [y],rdx
		jl .yloop
		
		add rsp, NUM_VARS*NBYTES		
		pop rbx
		pop rdi
		pop rsi
		ret
