;******************************************************************************/
;*                                                                            */
;*  Cedocida DV-Codec                                                         */
;*                                                                            */
;*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
;*  Project start: 15.12.2003                                                 */
;*  Last change  : 08.05.2005                                                 */
;*                                                                            */
;******************************************************************************/
;*                                                                            */
;*  This program is free software; you can redistribute it and/or modify      */
;*  it under the terms of the GNU General Public License as published by      */
;*  the Free Software Foundation; either version 2 of the License, or         */
;*  (at your option) any later version.                                       */
;*                                                                            */
;*  This program is distributed in the hope that it will be useful,           */
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
;*  GNU General Public License for more details.                              */
;*                                                                            */
;*  You should have received a copy of the GNU General Public License         */
;*  along with this program; if not, write to the Free Software               */
;*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
;*                                                                            */
;******************************************************************************/

%include "nasm.inc"

;=============================================================================
; Macros and other preprocessor constants
;=============================================================================


%define IDCT_248_MMX_COL_SHIFT 3
%define IDCT_248_MMX_ROW_SHIFT_1 (13+IDCT_248_MMX_COL_SHIFT)
%define IDCT_248_MMX_ROW_SHIFT_2 (20-IDCT_248_MMX_ROW_SHIFT_1)



%define BITS_INV_ACC    5                         ; 4 or 5 for IEEE
%define SHIFT_INV_ROW   (16 - BITS_INV_ACC)
%define SHIFT_INV_COL   (1 + BITS_INV_ACC)
%define RND_INV_ROW     (1024 * (6 - BITS_INV_ACC)) ; 1 << (SHIFT_INV_ROW-1)
%define RND_INV_COL     (16 * (BITS_INV_ACC - 3))   ; 1 << (SHIFT_INV_COL-1)
%define RND_INV_CORR    (RND_INV_COL - 1)           ; correction -1.0 and round

%define BITS_FRW_ACC    3                         ; 2 or 3 for accuracy
%define SHIFT_FRW_COL   BITS_FRW_ACC
%define SHIFT_FRW_ROW   (BITS_FRW_ACC + 17)
%define RND_FRW_ROW     (262144*(BITS_FRW_ACC - 1)) ; 1 << (SHIFT_FRW_ROW-1)




;=============================================================================
; Local Data (Read Only)
;=============================================================================

SECTION .rodata data align=16

;-----------------------------------------------------------------------------
; Various memory constants (trigonometric values or rounding values)
;-----------------------------------------------------------------------------

ALIGN 64
IDCT_248_table:
;fTab1:
  dw 0x4000, 0x4000, 0x58c5, 0x4b42
  dw 0x4000, 0x4000, 0x3249, 0x11a8
  dw 0x539f, 0x22a3, 0x4b42, 0xee58
  dw 0xdd5d, 0xac61, 0xa73b, 0xcdb7
  dw 0x4000, 0xc000, 0x3249, 0xa73b
  dw 0xc000, 0x4000, 0x11a8, 0x4b42
  dw 0x22a3, 0xac61, 0x11a8, 0xcdb7
  dw 0x539f, 0xdd5d, 0x4b42, 0xa73b

;fTab3:
  dw 0x539f, 0x539f, 0x73fc, 0x6254
  dw 0x539f, 0x539f, 0x41b3, 0x1712
  dw 0x6d41, 0x2d41, 0x6254, 0xe8ee
  dw 0xd2bf, 0x92bf, 0x8c04, 0xbe4d
  dw 0x539f, 0xac61, 0x41b3, 0x8c04
  dw 0xac61, 0x539f, 0x1712, 0x6254
  dw 0x2d41, 0x92bf, 0x1712, 0xbe4d
  dw 0x6d41, 0xd2bf, 0x6254, 0x8c04

IDCT_248_rounding:
  dw 8, 8, 8, 8, 6, 8, 8, 8
  dw 18,8, 8, 8, 8, 8, 8, 8

ALIGN 16
%define w_tan8 27146
pw_tan8:
  dw w_tan8,w_tan8,w_tan8,w_tan8

ALIGN 16
pw_one:
  dw 1, 1, 1, 1

ALIGN 16
one_corr:
  dw 1, 1, 1, 1
round_inv_row:
  dd RND_INV_ROW,  RND_INV_ROW
round_inv_col:
  dw RND_INV_COL,  RND_INV_COL,  RND_INV_COL, RND_INV_COL
round_inv_corr:
  dw RND_INV_CORR, RND_INV_CORR, RND_INV_CORR, RND_INV_CORR
round_frw_row:
  dd RND_FRW_ROW,  RND_FRW_ROW
tg_1_16:
  dw 13036,  13036,  13036,  13036    ; tg * (2<<16) + 0.5
tg_2_16:
  dw 27146,  27146,  27146,  27146    ; tg * (2<<16) + 0.5
tg_3_16:
  dw -21746, -21746, -21746, -21746    ; tg * (2<<16) + 0.5
cos_4_16:
  dw -19195, -19195, -19195, -19195    ; cos * (2<<16) + 0.5
ocos_4_16:
  dw 23170,  23170,  23170,  23170    ; cos * (2<<15) + 0.5
otg_3_16:
  dw 21895, 21895, 21895, 21895       ; tg * (2<<16) + 0.5

%if SHIFT_INV_ROW == 12   ; assume SHIFT_INV_ROW == 12
rounder_0:
  dd 65536, 65536
rounder_4:
  dd 0, 0
rounder_1:
  dd 7195, 7195
rounder_7
  dd 1024, 1024
rounder_2:
  dd 4520, 4520
rounder_6:
  dd 1024, 1024
rounder_3:
  dd 2407, 2407
rounder_5:
  dd 240, 240

%elif SHIFT_INV_ROW == 11   ; assume SHIFT_INV_ROW == 11
rounder_0:
  dd 65536, 65536
rounder_4:
  dd 0, 0
rounder_1:
  dd 3597, 3597
rounder_7:
  dd 512, 512
rounder_2:
  dd 2260, 2260
rounder_6:
  dd 512, 512
rounder_3:
  dd 1203, 1203
rounder_5:
  dd 120, 120
%else

%error invalid SHIFT_INV_ROW

%endif


;-----------------------------------------------------------------------------
; Tables for mmx processors
;-----------------------------------------------------------------------------

; Table for rows 0,4 - constants are multiplied by cos_4_16
ALIGN 64
tab_i_04_mmx:
  dw  16384,  16384,  16384, -16384    ; movq-> w06 w04 w02 w00
  dw  21407,   8867,   8867, -21407    ; w07 w05 w03 w01
  dw  16384, -16384,  16384,  16384    ; w14 w12 w10 w08
  dw  -8867,  21407, -21407,  -8867    ; w15 w13 w11 w09
  dw  22725,  12873,  19266, -22725    ; w22 w20 w18 w16
  dw  19266,   4520,  -4520, -12873    ; w23 w21 w19 w17
  dw  12873,   4520,   4520,  19266    ; w30 w28 w26 w24
  dw -22725,  19266, -12873, -22725    ; w31 w29 w27 w25

; Table for rows 1,7 - constants are multiplied by cos_1_16
tab_i_17_mmx:
  dw  22725,  22725,  22725, -22725    ; movq-> w06 w04 w02 w00
  dw  29692,  12299,  12299, -29692    ; w07 w05 w03 w01
  dw  22725, -22725,  22725,  22725    ; w14 w12 w10 w08
  dw -12299,  29692, -29692, -12299    ; w15 w13 w11 w09
  dw  31521,  17855,  26722, -31521    ; w22 w20 w18 w16
  dw  26722,   6270,  -6270, -17855    ; w23 w21 w19 w17
  dw  17855,   6270,   6270,  26722    ; w30 w28 w26 w24
  dw -31521,  26722, -17855, -31521    ; w31 w29 w27 w25

; Table for rows 2,6 - constants are multiplied by cos_2_16
tab_i_26_mmx:
  dw  21407,  21407,  21407, -21407    ; movq-> w06 w04 w02 w00
  dw  27969,  11585,  11585, -27969    ; w07 w05 w03 w01
  dw  21407, -21407,  21407,  21407    ; w14 w12 w10 w08
  dw -11585,  27969, -27969, -11585    ; w15 w13 w11 w09
  dw  29692,  16819,  25172, -29692    ; w22 w20 w18 w16
  dw  25172,   5906,  -5906, -16819    ; w23 w21 w19 w17
  dw  16819,   5906,   5906,  25172    ; w30 w28 w26 w24
  dw -29692,  25172, -16819, -29692    ; w31 w29 w27 w25

; Table for rows 3,5 - constants are multiplied by cos_3_16
tab_i_35_mmx:
  dw  19266,  19266,  19266, -19266    ; movq-> w06 w04 w02 w00
  dw  25172,  10426,  10426, -25172    ; w07 w05 w03 w01
  dw  19266, -19266,  19266,  19266    ; w14 w12 w10 w08
  dw -10426,  25172, -25172, -10426    ; w15 w13 w11 w09
  dw  26722,  15137,  22654, -26722    ; w22 w20 w18 w16
  dw  22654,   5315,  -5315, -15137    ; w23 w21 w19 w17
  dw  15137,   5315,   5315,  22654    ; w30 w28 w26 w24
  dw -26722,  22654, -15137, -26722    ; w31 w29 w27 w25


;=============================================================================
; Helper macros for the code
;=============================================================================

;-----------------------------------------------------------------------------
;
; The first stage iDCT 8x8 - inverse DCTs of rows
;
;-----------------------------------------------------------------------------
; The 8-point inverse DCT direct algorithm
;-----------------------------------------------------------------------------
;
; static const short w[32] = {
;	FIX(cos_4_16),  FIX(cos_2_16),  FIX(cos_4_16),  FIX(cos_6_16),
;	FIX(cos_4_16),  FIX(cos_6_16), -FIX(cos_4_16), -FIX(cos_2_16),
;	FIX(cos_4_16), -FIX(cos_6_16), -FIX(cos_4_16),  FIX(cos_2_16),
;	FIX(cos_4_16), -FIX(cos_2_16),  FIX(cos_4_16), -FIX(cos_6_16),
;	FIX(cos_1_16),  FIX(cos_3_16),  FIX(cos_5_16),  FIX(cos_7_16),
;	FIX(cos_3_16), -FIX(cos_7_16), -FIX(cos_1_16), -FIX(cos_5_16),
;	FIX(cos_5_16), -FIX(cos_1_16),  FIX(cos_7_16),  FIX(cos_3_16),
;	FIX(cos_7_16), -FIX(cos_5_16),  FIX(cos_3_16), -FIX(cos_1_16) };
;
; #define DCT_8_INV_ROW(x, y)
; {
; 	int a0, a1, a2, a3, b0, b1, b2, b3;
;
; 	a0 =x[0]*w[0]+x[2]*w[1]+x[4]*w[2]+x[6]*w[3];
; 	a1 =x[0]*w[4]+x[2]*w[5]+x[4]*w[6]+x[6]*w[7];
; 	a2 = x[0] * w[ 8] + x[2] * w[ 9] + x[4] * w[10] + x[6] * w[11];
; 	a3 = x[0] * w[12] + x[2] * w[13] + x[4] * w[14] + x[6] * w[15];
; 	b0 = x[1] * w[16] + x[3] * w[17] + x[5] * w[18] + x[7] * w[19];
; 	b1 = x[1] * w[20] + x[3] * w[21] + x[5] * w[22] + x[7] * w[23];
; 	b2 = x[1] * w[24] + x[3] * w[25] + x[5] * w[26] + x[7] * w[27];
; 	b3 = x[1] * w[28] + x[3] * w[29] + x[5] * w[30] + x[7] * w[31];
;
; 	y[0] = SHIFT_ROUND ( a0 + b0 );
; 	y[1] = SHIFT_ROUND ( a1 + b1 );
; 	y[2] = SHIFT_ROUND ( a2 + b2 );
; 	y[3] = SHIFT_ROUND ( a3 + b3 );
; 	y[4] = SHIFT_ROUND ( a3 - b3 );
; 	y[5] = SHIFT_ROUND ( a2 - b2 );
; 	y[6] = SHIFT_ROUND ( a1 - b1 );
; 	y[7] = SHIFT_ROUND ( a0 - b0 );
; }
;
;-----------------------------------------------------------------------------
;
; In this implementation the outputs of the iDCT-1D are multiplied
; 	for rows 0,4 - by cos_4_16,
; 	for rows 1,7 - by cos_1_16,
; 	for rows 2,6 - by cos_2_16,
; 	for rows 3,5 - by cos_3_16
; and are shifted to the left for better accuracy
;
; For the constants used,
; 	FIX(float_const) = (short) (float_const * (1<<15) + 0.5)
;
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; DCT_8_INV_ROW_MMX  INP, OUT, TABLE, ROUNDER
;-----------------------------------------------------------------------------
%macro DCT_8_INV_ROW_MMX 4
  movq mm0, [%1]            ; 0 ; x3 x2 x1 x0
  movq mm1, [%1+8]          ; 1 ; x7 x6 x5 x4
  movq mm2, mm0             ; 2  ; x3 x2 x1 x0
  movq mm3, [%3]            ; 3 ; w06 w04 w02 w00
  punpcklwd mm0, mm1        ; x5 x1 x4 x0
  movq mm5, mm0             ; 5 ; x5 x1 x4 x0
  punpckldq mm0, mm0        ; x4 x0 x4 x0
  movq mm4, [%3+8]          ; 4 ; w07 w05 w03 w01
  punpckhwd mm2, mm1        ; 1 ; x7 x3 x6 x2
  pmaddwd mm3, mm0          ; x4*w06+x0*w04 x4*w02+x0*w00
  movq mm6, mm2             ; 6 ; x7 x3 x6 x2
  movq mm1, [%3+32]         ; 1 ; w22 w20 w18 w16
  punpckldq mm2, mm2        ; x6 x2 x6 x2
  pmaddwd mm4, mm2          ; x6*w07+x2*w05 x6*w03+x2*w01
  punpckhdq mm5, mm5        ; x5 x1 x5 x1
  pmaddwd mm0, [%3+16]      ; x4*w14+x0*w12 x4*w10+x0*w08
  punpckhdq mm6, mm6        ; x7 x3 x7 x3
  movq mm7, [%3+40]         ; 7 ; w23 w21 w19 w17
  pmaddwd mm1, mm5          ; x5*w22+x1*w20 x5*w18+x1*w16
  paddd mm3, [%4]           ; +%4
  pmaddwd mm7, mm6          ; x7*w23+x3*w21 x7*w19+x3*w17
  pmaddwd mm2, [%3+24]      ; x6*w15+x2*w13 x6*w11+x2*w09
  paddd mm3, mm4            ; 4 ; a1=sum(even1) a0=sum(even0)
  pmaddwd mm5, [%3+48]      ; x5*w30+x1*w28 x5*w26+x1*w24
  movq mm4, mm3             ; 4 ; a1 a0
  pmaddwd mm6, [%3+56]      ; x7*w31+x3*w29 x7*w27+x3*w25
  paddd mm1, mm7            ; 7 ; b1=sum(odd1) b0=sum(odd0)
  paddd mm0, [%4]           ; +%4
  psubd mm3, mm1            ; a1-b1 a0-b0
  psrad mm3, SHIFT_INV_ROW  ; y6=a1-b1 y7=a0-b0
  paddd mm1, mm4            ; 4 ; a1+b1 a0+b0
  paddd mm0, mm2            ; 2 ; a3=sum(even3) a2=sum(even2)
  psrad mm1, SHIFT_INV_ROW  ; y1=a1+b1 y0=a0+b0
  paddd mm5, mm6            ; 6 ; b3=sum(odd3) b2=sum(odd2)
  movq mm4, mm0             ; 4 ; a3 a2
  paddd mm0, mm5            ; a3+b3 a2+b2
  psubd mm4, mm5            ; 5 ; a3-b3 a2-b2
  psrad mm0, SHIFT_INV_ROW  ; y3=a3+b3 y2=a2+b2
  psrad mm4, SHIFT_INV_ROW  ; y4=a3-b3 y5=a2-b2
  packssdw mm1, mm0         ; 0 ; y3 y2 y1 y0
  packssdw mm4, mm3         ; 3 ; y6 y7 y4 y5
  movq mm7, mm4             ; 7 ; y6 y7 y4 y5
  psrld mm4, 16             ; 0 y6 0 y4
  pslld mm7, 16             ; y7 0 y5 0
  movq [%2], mm1            ; 1 ; save y3 y2 y1 y0
  por mm7, mm4              ; 4 ; y7 y6 y5 y4
  movq [%2+8], mm7          ; 7 ; save y7 y6 y5 y4
%endmacro


;-----------------------------------------------------------------------------
; 4-point scaled IDCT algorithm
;-----------------------------------------------------------------------------
;
; --- some Matlab stuff (check idct-4 equation): ----
; w = tan(pi/8);
; KC10 = cos(pi/8);
; hu = [1/sqrt(8),1/2*KC10, 1/2, 1/2*KC10];
; kuz = [1,1,1,w; 1,w,-1,-1; 1,-w,-1,1; 1,-1,1,-w];
; x = [1,2,3,4];
; X = dct(x);
; y = (X.*hu)*(kuz')*sqrt(2)
;
; --- prescale (should be done at row-processing) ---
; x0 = x0 * 1/sqrt(8)
; x1 = x1 * 1/2*cos(pi/8)
; x2 = x2 * 1/2
; x3 = x3 * 1/2*cos(pi/8)
; x4 = x4 * 1/sqrt(8)
; x5 = x5 * 1/2*cos(pi/8)
; x6 = x6 * 1/2
; x7 = x7 * 1/2*cos(pi/8)
;
; --- step 1 ---
; r0 = x0 + x4
; r1 = x0 - x4
; r2 = x1 + x5
; r3 = x1 - x5
; r4 = x2 + x6
; r5 = x2 - x6
; r6 = x3 + x7
; r7 = x3 - x7
;
; --- step 2 ---
; s0 = r0 + r4
; s1 = r0 - r4
; s2 = r2 + r6*w
; s3 = r2*w - r6
; s4 = r1 + r5
; s5 = r1 - r5
; s6 = r3 + r7*w
; s7 = r3*w - r7
;
; --- step 3 ---
; y0 = s0 + s2
; y2 = s1 + s3
; y4 = s1 - s3
; y6 = s0 - s2
; y1 = s4 + s6
; y3 = s5 + s7
; y5 = s5 - s7
; y7 = s4 - s6


;-----------------------------------------------------------------------------
; IDCT_248_COL  INP,OUT
;-----------------------------------------------------------------------------
%macro IDCT_248_COL 2

%define x0 (%1 + 0*16)
%define x1 (%1 + 1*16)
%define x2 (%1 + 2*16)
%define x3 (%1 + 3*16)
%define x4 (%1 + 4*16)
%define x5 (%1 + 5*16)
%define x6 (%1 + 6*16)
%define x7 (%1 + 7*16)

%define y0 (%2 + 0*16)
%define y1 (%2 + 1*16)
%define y2 (%2 + 2*16)
%define y3 (%2 + 3*16)
%define y4 (%2 + 4*16)
%define y5 (%2 + 5*16)
%define y6 (%2 + 6*16)
%define y7 (%2 + 7*16)

; in and out can be the same adress!
ALIGN 16
;--- first the sum-part
	movq mm0, [x0]
	paddw mm0, [x4]		; r0 = x0 + x4

	movq mm1, [x2]
	paddw mm1, [x6]		; r4 = x2 + x6

	movq mm4, mm0		; r0
	paddw mm0, mm1		; s0 = r0 + r4
	movq mm7, [pw_tan8]	; w
	psubw mm4, mm1		; s1 = r0 - r4

	movq mm2, [x1]
	paddw mm2, [x5]		; r2 = x1 + x5

	movq mm3, [x3]
	paddw mm3, [x7]		; r6 = x3 + x7

	movq mm5, mm2		; r2
	pmulhw mm2, mm7		; r2*w
	
	movq mm6, mm4		; s1
	psubw mm2, mm3		; s3 = r2*w - r6
	
	pmulhw mm3, mm7		; r6*w
	paddw mm4, mm2		; y2 = s1 + s3

	paddw mm5, mm3		; s2 = r2 + r6*w
	psubw mm6, mm2		; y4 = s1 - s3

	movq mm1, mm0		; s0
	paddw mm0, mm5		; y0 = s0 + s2
	psubw mm1, mm5		; y6 = s0 - s2

;--- now the difference-part
	movq mm2, [x0]
	psubw mm2, [x4]		; r1 = x0 - x4

	movq mm5, [x2]
	psubw mm5, [x6]		; r5 = x2 - x6
 
	psraw mm4, SHIFT_INV_COL
	psraw mm6, SHIFT_INV_COL
	psraw mm0, SHIFT_INV_COL
	psraw mm1, SHIFT_INV_COL

	movq [y2], mm4
	movq [y4], mm6
	movq [y0], mm0
	movq [y6], mm1

	movq mm3, mm2		; r1
	paddw mm2, mm5		; s4 = r1 + r5
	psubw mm3, mm5		; s5 = r1 - r5

	movq mm0, [x1]
	psubw mm0, [x5]		; r3 = x1 - x5

	movq mm4, [x3]
	psubw mm4, [x7]		; r7 = x3 - x7

	movq mm6, mm0		; r3
	pmulhw mm0, mm7		; r3*w
	
	movq mm1, mm3		; s5
	psubw mm0, mm4		; s7 = r3*w - r7
	
	pmulhw mm4, mm7		; r7*w
	paddw mm3, mm0		; y3 = s5 + s7

	paddw mm6, mm4		; s6 = r3 + r7*w
	psubw mm1, mm0		; y5 = s5 - s7

	movq mm5, mm2		; s4
	paddw mm2, mm6		; y1 = s4 + s6
	psubw mm5, mm6		; y7 = s4 - s6

	psraw mm3, SHIFT_INV_COL
	psraw mm1, SHIFT_INV_COL
	psraw mm2, SHIFT_INV_COL
	psraw mm5, SHIFT_INV_COL

	movq [y3], mm3
	movq [y5], mm1
	movq [y1], mm2
	movq [y7], mm5

%endmacro









;=============================================================================
; Code
;=============================================================================

SECTION .text

cglobal idct_248_mmx_ad

;-----------------------------------------------------------------------------
; extern "C" void idct_248_mmx_ad( short *in, short *out );
;-----------------------------------------------------------------------------

ALIGN 16
idct_248_mmx_ad:

%ifndef WIN64
    mov rcx, dword [rsp + 4]
    mov rdx, dword [rsp + 8]
%endif    

	;; Process each row
    DCT_8_INV_ROW_MMX rcx+0*16, rdx+0*16, tab_i_04_mmx, rounder_0
    DCT_8_INV_ROW_MMX rcx+1*16, rdx+1*16, tab_i_26_mmx, rounder_2
    DCT_8_INV_ROW_MMX rcx+2*16, rdx+2*16, tab_i_04_mmx, rounder_4
    DCT_8_INV_ROW_MMX rcx+3*16, rdx+3*16, tab_i_26_mmx, rounder_2
    DCT_8_INV_ROW_MMX rcx+4*16, rdx+4*16, tab_i_04_mmx, rounder_0
    DCT_8_INV_ROW_MMX rcx+5*16, rdx+5*16, tab_i_26_mmx, rounder_2
    DCT_8_INV_ROW_MMX rcx+6*16, rdx+6*16, tab_i_04_mmx, rounder_4
    DCT_8_INV_ROW_MMX rcx+7*16, rdx+7*16, tab_i_26_mmx, rounder_2

	;; Process the columns (4 at a time)
    IDCT_248_COL rdx+0, rdx+0
    IDCT_248_COL rdx+8, rdx+8

    ret

