;******************************************************************************/
;*                                                                            */
;*  Cedocida DV-Codec                                                         */
;*                                                                            */
;*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
;*  Project start: 15.12.2003                                                 */
;*  Last change  : 08.05.2005                                                 */
;*                                                                            */
;******************************************************************************/
;*                                                                            */
;*  This program is free software; you can redistribute it and/or modify      */
;*  it under the terms of the GNU General Public License as published by      */
;*  the Free Software Foundation; either version 2 of the License, or         */
;*  (at your option) any later version.                                       */
;*                                                                            */
;*  This program is distributed in the hope that it will be useful,           */
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
;*  GNU General Public License for more details.                              */
;*                                                                            */
;*  You should have received a copy of the GNU General Public License         */
;*  along with this program; if not, write to the Free Software               */
;*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
;*                                                                            */
;******************************************************************************/

%ifndef NASM_INC_INCLUDED
%define NASM_INC_INCLUDED


%ifidn __OUTPUT_FORMAT__, win32
	%define WIN32
	%define NPTR 4
%elifidn __OUTPUT_FORMAT__, win64 
	%define WIN32 
	%define WIN64
	%define NPTR 8
	
	%ifdef PREFIX
		%undef PREFIX
		%warning "define PREFIX ignored for x64 build"
	%endif
	
	default rel
%else
	%error "only x86 and x64 builds supported!"
%endif

%macro cglobal 1
	%ifdef PREFIX
		global _%1
		%define %1 _%1
	%else
		global %1
	%endif
%endmacro

%macro cextern 1
	%ifdef PREFIX
		extern _%1
		%define %1 _%1
	%else
		extern %1
	%endif
%endmacro

%macro cfunction 1
	cglobal %1
	
	ALIGN 16
	%1:
%endmacro

%ifndef WIN64
	%define rax eax
	%define rbx ebx
	%define rcx ecx
	%define rdx edx
	%define rsi esi
	%define rdi edi
	%define rbp ebp
	%define rsp esp
	
	%define movsxd mov
	%define MOVSXD movsxd	
%endif
	
%macro INIT_SHADOWSTACK 0
%ifdef WIN64
	mov [rsp+1*8], rcx
	mov [rsp+2*8], rdx
	mov [rsp+3*8], r8
	mov [rsp+4*8], r9
%endif	
%endmacro


%macro ASSIGN_ARGS 2-5
%if %0 == 2
	ASSIGN_ARGS_1 %1, %2
%elif %0 == 3
	ASSIGN_ARGS_2 %1, %2, %3
%elif %0 == 4	
	ASSIGN_ARGS_3 %1, %2, %3, %4
%elif %0 == 5
	ASSIGN_ARGS_4 %1, %2, %3, %4, %5
%endif
%endmacro

%macro ASSIGN_ARGS_1 2
	%ifndef WIN64
	mov %1, [rsp+4*1+4*(%2)]
	%else
	mov %1, rcx
	%endif
%endmacro

%macro ASSIGN_ARGS_2 3
	%ifndef WIN64
	mov %1, [rsp+4*1+4*(%3)]
	mov %2, [rsp+4*2+4*(%3)]
	%else
	mov %1, rcx
	mov %2, rdx
	%endif
%endmacro

%macro ASSIGN_ARGS_3 4
	%ifndef WIN64
	mov %1, [rsp+4*1+4*(%4)]
	mov %2, [rsp+4*2+4*(%4)]
	mov %3, [rsp+4*3+4*(%4)]
	%else
	mov %1, rcx
	mov %2, rdx
	mov %3, r8
	%endif
%endmacro

%macro ASSIGN_ARGS_4 5
	%ifndef WIN64
	mov %1, [rsp+4*1+4*(%5)]
	mov %2, [rsp+4*2+4*(%5)]
	mov %3, [rsp+4*3+4*(%5)]
	mov %4, [rsp+4*4+4*(%5)]
	%else
	mov %1, rcx
	mov %2, rdx
	mov %3, r8
	mov %4, r9
	%endif
%endmacro


%endif