/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 08.05.2005                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include "../inttypes.h"
#include "../signalproc.h"

#include "bitstream.h"




void putbits( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b, const uint32_t code, const uint32_t length)
{
	if (length>32)
	{
		return;	// only for debugging!
	}

	if ( blk_bb->level <= 0 )
	{
		MB_VLC32b->pos[0] = code;
		MB_VLC32b->pos[1] = length;
		MB_VLC32b->pos += 2;
		return;
	}

	blk_bb->level = blk_bb->level - length;
	if ( blk_bb->level >= 0 )
	{
		uint32_t shift = 32 - ( blk_bb->modulo + length );

		if (shift<32)	// shift is unsigned!
		{
			if (shift!=0) 
			{
				blk_bb->pos[0] |= ( code << shift );
				blk_bb[0].modulo += length;
				return;
			}
			else
			{
				blk_bb->pos[0] |= code;
				blk_bb->pos++;
				blk_bb->modulo = 0;
			}
		} 
		else 
		{
			shift =  blk_bb->modulo + length - 32;
			blk_bb->pos[0] |= ( code >> shift );
			blk_bb->modulo = shift;
			shift = 32 - shift;
			blk_bb->pos[1] |= ( code << shift );
			blk_bb->pos++;
			return;
		}
	}
	else
	{
		blk_bb->level = blk_bb->level + length;

		uint32_t length_tmp = blk_bb->level;
		uint32_t shift = length-length_tmp;
		uint32_t code_tmp = code >> shift;

		if (length_tmp>32)
		{
			return;	// only for debugging!
		}
				
		putbits( blk_bb, MB_VLC32b, code_tmp, length_tmp);
		code_tmp = code & (~( 0xFFFFFFFF << shift ));

		
		MB_VLC32b->pos[0] = code_tmp;
		MB_VLC32b->pos[1] = shift;
		MB_VLC32b->pos += 2;
	}
		
	return;
}


int movebits( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	if ( MB_VLC32b->data != MB_VLC32b->pos )
	{
		uint32_t tmp[2];
		MB_VLC32buffer MB_VLC32b_tmp[1] = { { tmp, tmp } };

		while ( (MB_VLC32b->data < MB_VLC32b->pos) && (blk_bb->level > 0) )
		{
			uint32_t code = MB_VLC32b->data[0];
			uint32_t length = MB_VLC32b->data[1];
			putbits( blk_bb, MB_VLC32b_tmp, code, length );
			MB_VLC32b->data += 2;
		}

		if (MB_VLC32b_tmp->data != MB_VLC32b_tmp->pos)
		{
			MB_VLC32b->data -= 2;
			MB_VLC32b->data[0] = MB_VLC32b_tmp->data[0];
			MB_VLC32b->data[1] = MB_VLC32b_tmp->data[1];
		}

		return 0;
	}

	return 1;
}


void init_bitbuffer_encode( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b, uint32_t *MB_VLC32data, int VLC32dataSize,
						   uint8_t *DIF_block )
{
	uint32_t *data_ptr = ((uint32_t*)DIF_block);

	clear_block80( DIF_block, 5 );	

	int i=0;
	do
	{
		blk_bb[i+0].pos = data_ptr+1;
		blk_bb[i+0].level = 14*8;
		blk_bb[i+0].modulo = 0;

		blk_bb[i+1].pos = data_ptr+4;
		blk_bb[i+1].level = 14*8;
		blk_bb[i+1].modulo = 16;

		blk_bb[i+2].pos = data_ptr+8;
		blk_bb[i+2].level = 14*8;
		blk_bb[i+2].modulo = 0;

		blk_bb[i+3].pos = data_ptr+11;
		blk_bb[i+3].level = 14*8;
		blk_bb[i+3].modulo = 16;

		blk_bb[i+4].pos = data_ptr+15;
		blk_bb[i+4].level = 10*8;
		blk_bb[i+4].modulo = 0;

		blk_bb[i+5].pos = data_ptr+17;
		blk_bb[i+5].level = 10*8;
		blk_bb[i+5].modulo = 16;

		data_ptr+=20;
		i+=6;
	} while (i<30);

	i=0;
	do
	{
		MB_VLC32b[i].data = MB_VLC32data+i*VLC32dataSize;
		MB_VLC32b[i].pos = MB_VLC32data+i*VLC32dataSize;
		i++;
	} while (i<5);

}


void init_bitbuffer_decode( block_bitbuffer *blk_bb, uint8_t *DIF_block )
{
	uint32_t *data_ptr = ((uint32_t*)DIF_block);

	int i=0;
	do
	{
		blk_bb[i+0].pos = data_ptr+1;
		blk_bb[i+0].level = 14*8;
		blk_bb[i+0].modulo = 0;

		blk_bb[i+1].pos = data_ptr+4;
		blk_bb[i+1].level = 14*8;
		blk_bb[i+1].modulo = 16;

		blk_bb[i+2].pos = data_ptr+8;
		blk_bb[i+2].level = 14*8;
		blk_bb[i+2].modulo = 0;

		blk_bb[i+3].pos = data_ptr+11;
		blk_bb[i+3].level = 14*8;
		blk_bb[i+3].modulo = 16;

		blk_bb[i+4].pos = data_ptr+15;
		blk_bb[i+4].level = 10*8;
		blk_bb[i+4].modulo = 0;

		blk_bb[i+5].pos = data_ptr+17;
		blk_bb[i+5].level = 10*8;
		blk_bb[i+5].modulo = 16;

		data_ptr+=20;
		i+=6;
	} while (i<30);

}


int getbit( block_bitbuffer *blk_bb )
{
	uint32_t shift;
	uint32_t bit;

	if ( blk_bb->level <= 0 )
	{
		return -1;
	}
	else
	{
		blk_bb->level--;
		shift = 31-blk_bb->modulo;
		if ( shift!=0 )
		{
			bit = ( blk_bb->pos[0] >> shift )&0x01;
			blk_bb->modulo++;
			return bit;
		}
		else
		{
			bit = blk_bb->pos[0]&0x01;
			blk_bb->modulo = 0;
			blk_bb->pos++;
			return bit;
		}
	}
}

