/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include "../inttypes.h"

#include "../signalproc.h"
#include "bitstream.h"

#include "vlc.h"
#include "encode_vlc.h"
#include "encode_vlc_tables.h"




#define VLC32DATA_SIZE 64*6*2






#if !defined(_WIN64)
extern "C" {
	void block_scanning_88_unrolled_asm( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
	void block_scanning_248_unrolled_asm( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
	uint32_t get_vlc_length_88_unrolled_asm( short *block );
	uint32_t get_vlc_length_248_unrolled_asm( short *block );

	#define block_scanning_88_unrolled block_scanning_88_unrolled_asm
	#define block_scanning_248_unrolled block_scanning_248_unrolled_asm
	uint32_t (* const get_vlc_length_88_unrolled)( short *block ) = get_vlc_length_88_unrolled_asm;
	uint32_t (* const get_vlc_length_248_unrolled)( short *block ) = get_vlc_length_248_unrolled_asm;
}
#else
	static void block_scanning_88_unrolled_c( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
	static void block_scanning_248_unrolled_c( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b );
	static uint32_t get_vlc_length_c( const uint32_t run, int32_t amp);
	static uint32_t get_vlc_length_88_unrolled_c( int16_t *block );
	static uint32_t get_vlc_length_248_unrolled_c( int16_t *block );

	#define block_scanning_88_unrolled block_scanning_88_unrolled_c
	#define block_scanning_248_unrolled block_scanning_248_unrolled_c
	uint32_t (* const get_vlc_length_88_unrolled)( short *block ) = get_vlc_length_88_unrolled_c;
	uint32_t (* const get_vlc_length_248_unrolled)( short *block ) = get_vlc_length_248_unrolled_c;
#endif




extern "C" void put_vlc_code( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b, const uint32_t run, int32_t amp);



#ifdef _WIN64

static uint32_t get_vlc_length_c( const uint32_t run, int32_t amp)
{
	uint32_t length;

	if (amp<0) amp = -amp;
	if ((run)==0)
		length = (uint32_t)(vlc_table2[(amp)].length);
	else if (((amp)<=17) && ((run)<=4))
		length = vlc_table3[(run)*17+(amp)-18].length;
	else if (((amp)<=3) && ((run)<=14))
		length = vlc_table4[(run)*3+(amp)-16].length;
	else
	{
		length = (uint32_t)(vlc_table1[(run)-1].length);
		length = length + (uint32_t)(vlc_table2[(amp)].length);
	}
	return length;
}


#define GET_VLC_LENGTH_88_K( k ) \
	signed_amp = block[scan_order_88[k]];	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		length = length + get_vlc_length_c(run, signed_amp);	\
		run = -k;	\
	}

#define GET_VLC_LENGTH_248_K( k ) \
	signed_amp = block[scan_order_248[k]];	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		length = length + get_vlc_length_c( run, signed_amp);	\
		run = -k; 	\
	}

static uint32_t get_vlc_length_88_unrolled_c( int16_t *block )
{
	uint32_t length = 0;
	int32_t run = 0;
	int32_t signed_amp;

	GET_VLC_LENGTH_88_K(  1 )
	GET_VLC_LENGTH_88_K(  2 ) 
	GET_VLC_LENGTH_88_K(  3 ) 
	GET_VLC_LENGTH_88_K(  4 ) 
	GET_VLC_LENGTH_88_K(  5 ) 
	GET_VLC_LENGTH_88_K(  6 ) 
	GET_VLC_LENGTH_88_K(  7 ) 
	GET_VLC_LENGTH_88_K(  8 ) 
	GET_VLC_LENGTH_88_K(  9 ) 
	GET_VLC_LENGTH_88_K( 10 ) 
	GET_VLC_LENGTH_88_K( 11 ) 
	GET_VLC_LENGTH_88_K( 12 ) 
	GET_VLC_LENGTH_88_K( 13 ) 
	GET_VLC_LENGTH_88_K( 14 ) 
	GET_VLC_LENGTH_88_K( 15 ) 
	GET_VLC_LENGTH_88_K( 16 ) 
	GET_VLC_LENGTH_88_K( 17 ) 
	GET_VLC_LENGTH_88_K( 18 ) 
	GET_VLC_LENGTH_88_K( 19 ) 
	GET_VLC_LENGTH_88_K( 20 ) 
	GET_VLC_LENGTH_88_K( 21 ) 
	GET_VLC_LENGTH_88_K( 22 ) 
	GET_VLC_LENGTH_88_K( 23 ) 
	GET_VLC_LENGTH_88_K( 24 ) 
	GET_VLC_LENGTH_88_K( 25 ) 
	GET_VLC_LENGTH_88_K( 26 ) 
	GET_VLC_LENGTH_88_K( 27 ) 
	GET_VLC_LENGTH_88_K( 28 ) 
	GET_VLC_LENGTH_88_K( 29 ) 
	GET_VLC_LENGTH_88_K( 30 ) 
	GET_VLC_LENGTH_88_K( 31 ) 
	GET_VLC_LENGTH_88_K( 32 ) 
	GET_VLC_LENGTH_88_K( 33 ) 
	GET_VLC_LENGTH_88_K( 34 ) 
	GET_VLC_LENGTH_88_K( 35 ) 
	GET_VLC_LENGTH_88_K( 36 ) 
	GET_VLC_LENGTH_88_K( 37 ) 
	GET_VLC_LENGTH_88_K( 38 ) 
	GET_VLC_LENGTH_88_K( 39 ) 
	GET_VLC_LENGTH_88_K( 40 ) 
	GET_VLC_LENGTH_88_K( 41 ) 
	GET_VLC_LENGTH_88_K( 42 ) 
	GET_VLC_LENGTH_88_K( 43 ) 
	GET_VLC_LENGTH_88_K( 44 ) 
	GET_VLC_LENGTH_88_K( 45 ) 
	GET_VLC_LENGTH_88_K( 46 ) 

	if (( (uint32_t)block[102/2] |
		 *(uint32_t*)&(block[ 76/2]) |
		 *(uint32_t*)&(block[ 92/2]) |
		 *(uint32_t*)&(block[104/2]) |
		 *(uint32_t*)&(block[108/2]) |
		 *(uint32_t*)&(block[116/2]) |
		 *(uint32_t*)&(block[120/2]) |
		 *(uint32_t*)&(block[124/2]) |
		 (uint32_t)block[ 62/2] |
		 (uint32_t)block[ 90/2] ) == 0)
	{
		return length + 4;	//EOB
	}

	GET_VLC_LENGTH_88_K( 47 ) 
	GET_VLC_LENGTH_88_K( 48 ) 
	GET_VLC_LENGTH_88_K( 49 ) 
	GET_VLC_LENGTH_88_K( 50 ) 
	GET_VLC_LENGTH_88_K( 51 ) 
	GET_VLC_LENGTH_88_K( 52 ) 
	GET_VLC_LENGTH_88_K( 53 ) 
	GET_VLC_LENGTH_88_K( 54 ) 
	GET_VLC_LENGTH_88_K( 55 ) 
	GET_VLC_LENGTH_88_K( 56 ) 
	GET_VLC_LENGTH_88_K( 57 ) 
	GET_VLC_LENGTH_88_K( 58 ) 
	GET_VLC_LENGTH_88_K( 59 ) 
	GET_VLC_LENGTH_88_K( 60 ) 
	GET_VLC_LENGTH_88_K( 61 ) 
	GET_VLC_LENGTH_88_K( 62 ) 
	GET_VLC_LENGTH_88_K( 63 )	

	//if (run!=(-63)) length = length + 4;	//EOB
	return length + 4;	//EOB
}


static uint32_t get_vlc_length_248_unrolled_c( int16_t *block )
{
	uint32_t length = 0;
	int32_t run = 0;
	int32_t signed_amp;

	GET_VLC_LENGTH_248_K(  1 )
	GET_VLC_LENGTH_248_K(  2 ) 
	GET_VLC_LENGTH_248_K(  3 ) 
	GET_VLC_LENGTH_248_K(  4 ) 
	GET_VLC_LENGTH_248_K(  5 ) 
	GET_VLC_LENGTH_248_K(  6 ) 
	GET_VLC_LENGTH_248_K(  7 ) 
	GET_VLC_LENGTH_248_K(  8 ) 
	GET_VLC_LENGTH_248_K(  9 ) 
	GET_VLC_LENGTH_248_K( 10 ) 
	GET_VLC_LENGTH_248_K( 11 ) 
	GET_VLC_LENGTH_248_K( 12 ) 
	GET_VLC_LENGTH_248_K( 13 ) 
	GET_VLC_LENGTH_248_K( 14 ) 
	GET_VLC_LENGTH_248_K( 15 ) 
	GET_VLC_LENGTH_248_K( 16 ) 
	GET_VLC_LENGTH_248_K( 17 ) 
	GET_VLC_LENGTH_248_K( 18 ) 
	GET_VLC_LENGTH_248_K( 19 ) 
	GET_VLC_LENGTH_248_K( 20 ) 
	GET_VLC_LENGTH_248_K( 21 ) 
	GET_VLC_LENGTH_248_K( 22 ) 
	GET_VLC_LENGTH_248_K( 23 ) 
	GET_VLC_LENGTH_248_K( 24 ) 
	GET_VLC_LENGTH_248_K( 25 ) 
	GET_VLC_LENGTH_248_K( 26 ) 
	GET_VLC_LENGTH_248_K( 27 ) 
	GET_VLC_LENGTH_248_K( 28 ) 
	GET_VLC_LENGTH_248_K( 29 ) 
	GET_VLC_LENGTH_248_K( 30 ) 
	GET_VLC_LENGTH_248_K( 31 ) 
	GET_VLC_LENGTH_248_K( 32 ) 
	GET_VLC_LENGTH_248_K( 33 ) 
	GET_VLC_LENGTH_248_K( 34 ) 
	GET_VLC_LENGTH_248_K( 35 ) 
	GET_VLC_LENGTH_248_K( 36 ) 
	GET_VLC_LENGTH_248_K( 37 ) 
	GET_VLC_LENGTH_248_K( 38 ) 
	GET_VLC_LENGTH_248_K( 39 ) 
	GET_VLC_LENGTH_248_K( 40 ) 
	GET_VLC_LENGTH_248_K( 41 ) 
	GET_VLC_LENGTH_248_K( 42 ) 
	GET_VLC_LENGTH_248_K( 43 ) 
	GET_VLC_LENGTH_248_K( 44 ) 
	GET_VLC_LENGTH_248_K( 45 ) 
	GET_VLC_LENGTH_248_K( 46 ) 
	GET_VLC_LENGTH_248_K( 47 ) 
	GET_VLC_LENGTH_248_K( 48 ) 
	GET_VLC_LENGTH_248_K( 49 ) 
	GET_VLC_LENGTH_248_K( 50 ) 
	GET_VLC_LENGTH_248_K( 51 ) 
	GET_VLC_LENGTH_248_K( 52 ) 
	GET_VLC_LENGTH_248_K( 53 ) 
	GET_VLC_LENGTH_248_K( 54 ) 
	GET_VLC_LENGTH_248_K( 55 ) 
	GET_VLC_LENGTH_248_K( 56 ) 
	GET_VLC_LENGTH_248_K( 57 ) 
	GET_VLC_LENGTH_248_K( 58 ) 
	GET_VLC_LENGTH_248_K( 59 ) 
	GET_VLC_LENGTH_248_K( 60 ) 
	GET_VLC_LENGTH_248_K( 61 ) 
	GET_VLC_LENGTH_248_K( 62 ) 
	GET_VLC_LENGTH_248_K( 63 )	

	//if (run!=(-63)) length = length + 4;	//EOB
	length = length + 4;	//EOB

	return length;
}


#define BLOCK_SCANNING_88_K( k ) \
	signed_amp = (int32_t)( block[scan_order_88[k]] );	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		put_vlc_code( blk_bb, MB_VLC32b, run, signed_amp);	\
		run = -k;	\
	}

#define BLOCK_SCANNING_248_K( k ) \
	signed_amp = (int32_t)( block[scan_order_248[k]] );	\
	if (signed_amp!=0)	\
	{	\
		run = run + (k-1);	\
		put_vlc_code( blk_bb, MB_VLC32b, run, signed_amp);	\
		run = -k;	\
	}

static void block_scanning_88_unrolled_c( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	int32_t run = 0;
	int32_t signed_amp;

	BLOCK_SCANNING_88_K(  1 ) 
	BLOCK_SCANNING_88_K(  2 ) 
	BLOCK_SCANNING_88_K(  3 ) 
	BLOCK_SCANNING_88_K(  4 ) 
	BLOCK_SCANNING_88_K(  5 ) 
	BLOCK_SCANNING_88_K(  6 ) 
	BLOCK_SCANNING_88_K(  7 ) 
	BLOCK_SCANNING_88_K(  8 ) 
	BLOCK_SCANNING_88_K(  9 ) 
	BLOCK_SCANNING_88_K( 10 ) 
	BLOCK_SCANNING_88_K( 11 ) 
	BLOCK_SCANNING_88_K( 12 ) 
	BLOCK_SCANNING_88_K( 13 ) 
	BLOCK_SCANNING_88_K( 14 ) 
	BLOCK_SCANNING_88_K( 15 ) 
	BLOCK_SCANNING_88_K( 16 ) 
	BLOCK_SCANNING_88_K( 17 ) 
	BLOCK_SCANNING_88_K( 18 ) 
	BLOCK_SCANNING_88_K( 19 ) 
	BLOCK_SCANNING_88_K( 20 ) 
	BLOCK_SCANNING_88_K( 21 ) 
	BLOCK_SCANNING_88_K( 22 ) 
	BLOCK_SCANNING_88_K( 23 ) 
	BLOCK_SCANNING_88_K( 24 ) 
	BLOCK_SCANNING_88_K( 25 ) 
	BLOCK_SCANNING_88_K( 26 ) 
	BLOCK_SCANNING_88_K( 27 ) 
	BLOCK_SCANNING_88_K( 28 ) 
	BLOCK_SCANNING_88_K( 29 ) 
	BLOCK_SCANNING_88_K( 30 ) 
	BLOCK_SCANNING_88_K( 31 ) 
	BLOCK_SCANNING_88_K( 32 ) 
	BLOCK_SCANNING_88_K( 33 ) 
	BLOCK_SCANNING_88_K( 34 ) 
	BLOCK_SCANNING_88_K( 35 ) 
	BLOCK_SCANNING_88_K( 36 ) 
	BLOCK_SCANNING_88_K( 37 ) 
	BLOCK_SCANNING_88_K( 38 ) 
	BLOCK_SCANNING_88_K( 39 ) 
	BLOCK_SCANNING_88_K( 40 ) 
	BLOCK_SCANNING_88_K( 41 ) 
	BLOCK_SCANNING_88_K( 42 ) 
	BLOCK_SCANNING_88_K( 43 ) 
	BLOCK_SCANNING_88_K( 44 ) 
	BLOCK_SCANNING_88_K( 45 ) 
	BLOCK_SCANNING_88_K( 46 ) 

	if (( (uint32_t)block[102/2] |
		 *(uint32_t*)&(block[ 76/2]) |
		 *(uint32_t*)&(block[ 92/2]) |
		 *(uint32_t*)&(block[104/2]) |
		 *(uint32_t*)&(block[108/2]) |
		 *(uint32_t*)&(block[116/2]) |
		 *(uint32_t*)&(block[120/2]) |
		 *(uint32_t*)&(block[124/2]) |
		 (uint32_t)block[ 62/2] |
		 (uint32_t)block[ 90/2] ) == 0)
	{
		putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
		return;
	}

	BLOCK_SCANNING_88_K( 47 ) 
	BLOCK_SCANNING_88_K( 48 ) 
	BLOCK_SCANNING_88_K( 49 ) 
	BLOCK_SCANNING_88_K( 50 ) 
	BLOCK_SCANNING_88_K( 51 ) 
	BLOCK_SCANNING_88_K( 52 ) 
	BLOCK_SCANNING_88_K( 53 ) 
	BLOCK_SCANNING_88_K( 54 ) 
	BLOCK_SCANNING_88_K( 55 ) 
	BLOCK_SCANNING_88_K( 56 ) 
	BLOCK_SCANNING_88_K( 57 ) 
	BLOCK_SCANNING_88_K( 58 ) 
	BLOCK_SCANNING_88_K( 59 ) 
	BLOCK_SCANNING_88_K( 60 ) 
	BLOCK_SCANNING_88_K( 61 ) 
	BLOCK_SCANNING_88_K( 62 ) 
	BLOCK_SCANNING_88_K( 63 )	

	//if (run!=(-63)) putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
	putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
}


static void block_scanning_248_unrolled_c( int16_t *block, block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b )
{
	int32_t run = 0;
	int32_t signed_amp;

	BLOCK_SCANNING_248_K(  1 ) 
	BLOCK_SCANNING_248_K(  2 ) 
	BLOCK_SCANNING_248_K(  3 ) 
	BLOCK_SCANNING_248_K(  4 ) 
	BLOCK_SCANNING_248_K(  5 ) 
	BLOCK_SCANNING_248_K(  6 ) 
	BLOCK_SCANNING_248_K(  7 ) 
	BLOCK_SCANNING_248_K(  8 ) 
	BLOCK_SCANNING_248_K(  9 ) 
	BLOCK_SCANNING_248_K( 10 ) 
	BLOCK_SCANNING_248_K( 11 ) 
	BLOCK_SCANNING_248_K( 12 ) 
	BLOCK_SCANNING_248_K( 13 ) 
	BLOCK_SCANNING_248_K( 14 ) 
	BLOCK_SCANNING_248_K( 15 ) 
	BLOCK_SCANNING_248_K( 16 ) 
	BLOCK_SCANNING_248_K( 17 ) 
	BLOCK_SCANNING_248_K( 18 ) 
	BLOCK_SCANNING_248_K( 19 ) 
	BLOCK_SCANNING_248_K( 20 ) 
	BLOCK_SCANNING_248_K( 21 ) 
	BLOCK_SCANNING_248_K( 22 ) 
	BLOCK_SCANNING_248_K( 23 ) 
	BLOCK_SCANNING_248_K( 24 ) 
	BLOCK_SCANNING_248_K( 25 ) 
	BLOCK_SCANNING_248_K( 26 ) 
	BLOCK_SCANNING_248_K( 27 ) 
	BLOCK_SCANNING_248_K( 28 ) 
	BLOCK_SCANNING_248_K( 29 ) 
	BLOCK_SCANNING_248_K( 30 ) 
	BLOCK_SCANNING_248_K( 31 ) 
	BLOCK_SCANNING_248_K( 32 ) 
	BLOCK_SCANNING_248_K( 33 ) 
	BLOCK_SCANNING_248_K( 34 ) 
	BLOCK_SCANNING_248_K( 35 ) 
	BLOCK_SCANNING_248_K( 36 ) 
	BLOCK_SCANNING_248_K( 37 ) 
	BLOCK_SCANNING_248_K( 38 ) 
	BLOCK_SCANNING_248_K( 39 ) 
	BLOCK_SCANNING_248_K( 40 ) 
	BLOCK_SCANNING_248_K( 41 ) 
	BLOCK_SCANNING_248_K( 42 ) 
	BLOCK_SCANNING_248_K( 43 ) 
	BLOCK_SCANNING_248_K( 44 ) 
	BLOCK_SCANNING_248_K( 45 ) 
	BLOCK_SCANNING_248_K( 46 ) 
	BLOCK_SCANNING_248_K( 47 ) 
	BLOCK_SCANNING_248_K( 48 ) 
	BLOCK_SCANNING_248_K( 49 ) 
	BLOCK_SCANNING_248_K( 50 ) 
	BLOCK_SCANNING_248_K( 51 ) 
	BLOCK_SCANNING_248_K( 52 ) 
	BLOCK_SCANNING_248_K( 53 ) 
	BLOCK_SCANNING_248_K( 54 ) 
	BLOCK_SCANNING_248_K( 55 ) 
	BLOCK_SCANNING_248_K( 56 ) 
	BLOCK_SCANNING_248_K( 57 ) 
	BLOCK_SCANNING_248_K( 58 ) 
	BLOCK_SCANNING_248_K( 59 ) 
	BLOCK_SCANNING_248_K( 60 ) 
	BLOCK_SCANNING_248_K( 61 ) 
	BLOCK_SCANNING_248_K( 62 ) 
	BLOCK_SCANNING_248_K( 63 )

	//if (run!=(-63)) putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
	putbits( blk_bb, MB_VLC32b, 0x06, 4 );	//EOB
}

#endif



void put_vlc_code( block_bitbuffer *blk_bb, MB_VLC32buffer *MB_VLC32b, const uint32_t run, int32_t amp)
{
	if (amp>=0)
	{
		if ((run)==0) 	
		{	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code);	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=17) && ((run)<=4))	
		{	
			uint32_t code = vlc_table3[(run)*17+(amp)-18].code;	
			uint32_t length = vlc_table3[(run)*17+(amp)-18].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=3) && ((run)<=14))	
		{	
			uint32_t code = vlc_table4[(run)*3+(amp)-16].code;	
			uint32_t length = vlc_table4[(run)*3+(amp)-16].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else	
		{	
			uint32_t code1 = (uint32_t)(vlc_table1[(run)-1].code);	
			uint32_t length1 = (uint32_t)(vlc_table1[(run)-1].length);	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code);	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			code = ( (code1<<length) | code );	
			length = length+length1;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
	}
	else
	{
		amp = -amp;
		if ((run)==0) 	
		{	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code) + 1;	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=17) && ((run)<=4))	
		{	
			uint32_t code = vlc_table3[(run)*17+(amp)-18].code + 1;	
			uint32_t length = vlc_table3[(run)*17+(amp)-18].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else if (((amp)<=3) && ((run)<=14))	
		{	
			uint32_t code = vlc_table4[(run)*3+(amp)-16].code + 1;	
			uint32_t length = vlc_table4[(run)*3+(amp)-16].length;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}	
		else	
		{	
			uint32_t code1 = (uint32_t)(vlc_table1[(run)-1].code);	
			uint32_t length1 = (uint32_t)(vlc_table1[(run)-1].length);	
			uint32_t code = (uint32_t)(vlc_table2[(amp)].code + 1);	
			uint32_t length = (uint32_t)(vlc_table2[(amp)].length);	
			code = ( (code1<<length) | code );	
			length = length+length1;	
			putbits( blk_bb, MB_VLC32b, code, length );	
		}
	}
}


#define VLC32DATA_SIZE 64*6*2

int encode_videosegment( int16_t *QTMB, uint32_t *quantization_number, uint32_t *DCT_mode, uint32_t *class_number, 
						   uint8_t *DIF_block, uint32_t DIF_sequence_number, uint32_t DIF_block_number )
{
	uint32_t MB_VLC32data[VLC32DATA_SIZE*5];	
	block_bitbuffer blk_bb[5*6];
	MB_VLC32buffer MB_VLC32b[5];

	int remain;
	int j;
	int n = 0;
	int16_t *block = QTMB;

	init_bitbuffer_encode( blk_bb, MB_VLC32b, MB_VLC32data, VLC32DATA_SIZE, DIF_block );


	for (j=0; j<5; j++)
	{
		int l;
		
		const uint32_t SCT = 4;	// DIF-block-type: video
		const uint32_t Seq = 15;	// Video-sequenznumber: not available
		uint32_t Dseq = DIF_sequence_number;
		uint32_t DBN = DIF_block_number + j;
		const uint32_t STA = 0;	// Macroblock-state
		uint32_t QNO = quantization_number[j];

		((uint32_t*)DIF_block)[j*20] =  (SCT<<29) | 0x10070000 | (Seq<<24) | (Dseq<<20) | (DBN<<8) | (STA<<4) | QNO ;

		for (l=0;l<6;l++)
		{
			uint32_t mo = DCT_mode[n];
			uint32_t c1c0 = class_number[n];
			uint32_t DCI = ( block[0] & 0x1FF );
			uint32_t DCI_m0_c1_c0 = ( (DCI<<3) | (mo<<2) | c1c0 );

			putbits( blk_bb+n, MB_VLC32b+j, DCI_m0_c1_c0, 12 );

			if ( mo==0 )
			{
				block_scanning_88_unrolled( block, blk_bb+n, MB_VLC32b+j );
			}
			else
			{
				block_scanning_248_unrolled( block, blk_bb+n, MB_VLC32b+j );
			}

			block += 64;
			n++;
		}

		// distribute "overflowed" bits on blocks
		if ( movebits( blk_bb+n-6, MB_VLC32b+j ) == 0)
		{
			movebits( blk_bb+n-5, MB_VLC32b+j );
			movebits( blk_bb+n-4, MB_VLC32b+j );
			movebits( blk_bb+n-3, MB_VLC32b+j );
			movebits( blk_bb+n-2, MB_VLC32b+j );
			movebits( blk_bb+n-1, MB_VLC32b+j ); 
		}

	}

	// distribute "overflowed" bits on blocks
	for (j=0; j<5; j++)
	{
		int l;
		n = 0;
		for (l=0; l<4; l++)
		{
			if (j==l) n+=6;
			if( movebits( blk_bb+n+0, MB_VLC32b+j )==1 ) break;
			movebits( blk_bb+n+1, MB_VLC32b+j );
			movebits( blk_bb+n+2, MB_VLC32b+j );
			movebits( blk_bb+n+3, MB_VLC32b+j );
			movebits( blk_bb+n+4, MB_VLC32b+j );
			movebits( blk_bb+n+5, MB_VLC32b+j );
			n+=6;
		}
	}

	byte_swap_block80( DIF_block, 5 );

	remain = 0;
	remain = remain + (int)(MB_VLC32b[0].pos - MB_VLC32b[0].data);
	remain = remain + (int)(MB_VLC32b[1].pos - MB_VLC32b[1].data);
	remain = remain + (int)(MB_VLC32b[2].pos - MB_VLC32b[2].data);
	remain = remain + (int)(MB_VLC32b[3].pos - MB_VLC32b[3].data);
	remain = remain + (int)(MB_VLC32b[4].pos - MB_VLC32b[4].data);

	return (remain>>1);
}


uint32_t get_videosegment_length( int16_t *QTMB, uint32_t *DCT_mode )
{
	uint32_t vs_length_bits = 5*4*8 + 5*6*12;
	int j;
	int16_t *block = QTMB;

	for (j=0; j<(5*6); j++)
	{
		uint32_t mo = DCT_mode[j];

		if ( mo==0 )
		{
			vs_length_bits += get_vlc_length_88_unrolled( block );
		}
		else
		{
			vs_length_bits += get_vlc_length_248_unrolled( block );
		}

		if (vs_length_bits>(5*67*8)) return (vs_length_bits);

		block += 64;
	}

	return (vs_length_bits);
}




